

import configparser
import copy
import os
import locale

from gi.repository import GLib, GObject

from kupfer import config, pretty, scheduler


def strbool(value, default=False):
    """Coerce bool from string value or bool"""
    if value in (True, False):
        return value
    value = str(value).lower()
    if value in ("no", "false"):
        return False
    if value in ("yes", "true"):
        return True
    return default

def strint(value, default=0):
    """Coerce bool from string value or bool"""
    try:
        return int(value)
    except ValueError:
        return default

def _override_encoding(name):
    """
    Return a new encoding name if we want to override it, else return None.

    This is used to “upgrade” ascii to UTF-8 since the latter is a superset.
    """
    if name.lower() in ("ascii", "ANSI_X3.4-1968".lower()):
        return "UTF-8"
    else:
        return None

class SettingsController (GObject.GObject, pretty.OutputMixin):
    __gtype_name__ = "SettingsController"
    config_filename = "kupfer.cfg"
    defaults_filename = "defaults.cfg"
    sep = ";"
    default_directories = ("~/", "~/Desktop", )
    # Minimal "defaults" to define all fields
    # Read defaults defined in a defaults.cfg file
    defaults = {
        "Kupfer": {
            "keybinding" : "" ,
            "magickeybinding": "",
            "showstatusicon" : True,
            "showstatusicon_ai" : False,
            "usecommandkeys" : True,
        },
        "Appearance": {
            "icon_large_size": 128,
            "icon_small_size": 24,
            "list_height": 200,
        },
        "Directories" : { "direct" : default_directories, "catalog" : (), },
        "DeepDirectories" : { "direct" : (), "catalog" : (), "depth" : 1, },
        'Keybindings': {},
        "Tools": {},
    }
    def __init__(self):
        GObject.GObject.__init__(self)
        self._defaults_path = None
        self.encoding = _override_encoding(locale.getpreferredencoding())
        self.output_debug("Using", self.encoding)
        self._config = self._read_config()
        self._save_timer = scheduler.Timer(True)
        self._alternatives = {}
        self._alternative_validators = {}

    def _update_config_save_timer(self):
        self._save_timer.set(60, self._save_config)

    def _read_config(self, read_config=True):
        """
        Read cascading config files
        default -> then config
        (in all XDG_CONFIG_DIRS)
        """
        parser = configparser.RawConfigParser()

        def fill_parser(parser, defaults):
            for secname, section in defaults.items():
                if not parser.has_section(secname):
                    parser.add_section(secname)
                for key, default in section.items():
                    if isinstance(default, (tuple, list)):
                        default = self.sep.join(default)
                    elif isinstance(default, int):
                        default = str(default)
                    parser.set(secname, key, default)

        # Set up defaults
        confmap = copy.deepcopy(self.defaults)
        fill_parser(parser, confmap)

        # Read all config files
        config_files = []
        try:
            defaults_path = config.get_data_file(self.defaults_filename)
        except config.ResourceLookupError:
            self.output_error(("Error: no default config file %s found!" % self.defaults_filename))
        else:
            self._defaults_path = defaults_path
            config_files += (defaults_path, )

        if read_config:
            config_path = config.get_config_file(self.config_filename)
            if config_path:
                config_files += (config_path, )

        for config_file in config_files:
            try:
                parser.read(config_file, encoding=self.encoding)
            except IOError as e:
                self.output_error(("Error reading configuration file %s: %s" % (config_file, e)))
            except UnicodeDecodeError as e:
                self.output_error(("Error reading configuration file %s: %s" % (config_file, e)))

        # Read parsed file into the dictionary again
        for secname in parser.sections():
            if secname not in confmap: confmap[secname] = {}
            for key in parser.options(secname):
                value = parser.get(secname, key)
                retval = value
                if secname in self.defaults and key in self.defaults[secname]:
                    defval = self.defaults[secname][key]
                    if isinstance(defval, (tuple, list)):
                        if not value:
                            retval = ()
                        else:
                            retval = [p.strip() for p in value.split(self.sep) if p]
                    elif isinstance(defval, bool):
                        retval = strbool(value)
                    elif isinstance(defval, int):
                        retval = type(defval)(value)
                    else:
                        retval = str(value)
                confmap[secname][key] = retval

        return confmap

    def _save_config(self, scheduler=None):
        self.output_debug("Saving config")
        config_path = config.save_config_file(self.config_filename)
        if not config_path:
            self.output_info("Unable to save settings, can't find config dir")
            return
        # read in just the default values
        default_confmap = self._read_config(read_config=False)

        def confmap_difference(config, defaults):
            """Extract the non-default keys to write out"""
            difference = dict()
            for secname, section in config.items():
                if secname not in defaults:
                    difference[secname] = dict(section)
                    continue
                difference[secname] = {}
                for key, config_val in section.items():
                    if (secname in defaults and
                            key in defaults[secname]):
                        if defaults[secname][key] == config_val:
                            continue
                    difference[secname][key] = config_val
                if not difference[secname]:
                    del difference[secname]
            return difference

        parser = configparser.RawConfigParser()
        def fill_parser(parser, defaults):
            for secname in sorted(defaults):
                section = defaults[secname]
                if not parser.has_section(secname):
                    parser.add_section(secname)
                for key in sorted(section):
                    value = section[key]
                    if isinstance(value, (tuple, list)):
                        value = self.sep.join(value)
                    elif isinstance(value, int):
                        value = str(value)
                    parser.set(secname, key, value)

        confmap = confmap_difference(self._config, default_confmap)
        fill_parser(parser, confmap)
        ## Write to tmp then rename over for it to be atomic
        temp_config_path = "%s.%s" % (config_path, os.getpid())
        with open(temp_config_path, "w") as out:
            parser.write(out)
        os.rename(temp_config_path, config_path)

    def get_config(self, section, key):
        """General interface, but section must exist"""
        key = key.lower()
        value = self._config[section].get(key)
        if section in self.defaults:
            return value
        raise KeyError("Invalid settings section: %s" % section)

    def _set_config(self, section, key, value):
        """General interface, but section must exist"""
        self.output_debug("Set", section, key, "to", value)
        key = key.lower()
        oldvalue = self._config[section].get(key)
        if section in self.defaults:
            value_type = type(oldvalue) if oldvalue is not None else str
            self._config[section][key] = value_type(value)
            self._emit_value_changed(section, key, value)
            self._update_config_save_timer()
            return True
        raise KeyError("Invalid settings section: %s" % section)

    def _emit_value_changed(self, section, key, value):
        suffix = "%s.%s" % (section.lower(), key.lower())
        self.emit("value-changed::"+suffix, section, key, value)

    def _get_raw_config(self, section, key):
        """General interface, but section must exist"""
        key = key.lower()
        value = self._config[section].get(key)
        return value

    def _set_raw_config(self, section, key, value):
        """General interface, but will create section"""
        self.output_debug("Set", section, key, "to", value)
        key = key.lower()
        if section not in self._config:
            self._config[section] = {}
        self._config[section][key] = str(value)
        self._update_config_save_timer()
        return False

    def get_from_defaults(self, section, option=None):
        """Load values from default configuration file.
        If @option is None, return all section items as (key, value) """
        if self._defaults_path is None:
            self.output_error('Defaults not found')
            return
        parser = configparser.RawConfigParser()
        parser.read(self._defaults_path)
        if option is None:
            return parser.items(section)
        else:
            return parser.get(section, option.lower())

    def get_config_int(self, section, key):
        """section must exist"""
        key = key.lower()
        value = self._config[section].get(key)
        if section in self.defaults:
            return strint(value)
        raise KeyError("Invalid settings section: %s" % section)


    def get_plugin_enabled(self, plugin_id):
        """Convenience: if @plugin_id is enabled"""
        return self.get_plugin_config(plugin_id, "kupfer_enabled",
                value_type=strbool, default=False)

    def set_plugin_enabled(self, plugin_id, enabled):
        """Convenience: set if @plugin_id is enabled"""
        ret = self.set_plugin_config(plugin_id, "kupfer_enabled", enabled,
                value_type=strbool)
        self.emit("plugin-enabled-changed", plugin_id, enabled)
        return ret

    def get_plugin_is_hidden(self, plugin_id):
        """Convenience: if @plugin_id is hidden"""
        return self.get_plugin_config(plugin_id, "kupfer_hidden",
                value_type=strbool, default=False)

    @classmethod
    def _source_config_repr(self, obj):
        name = type(obj).__name__
        return "".join([(c if c.isalnum() else '_') for c in name])

    def get_source_is_toplevel(self, plugin_id, src):
        key = "kupfer_toplevel_" + self._source_config_repr(src)
        default = not getattr(src, "source_prefer_sublevel", False)
        return self.get_plugin_config(plugin_id, key,
                                      value_type=strbool, default=default)

    def set_source_is_toplevel(self, plugin_id, src, value):
        key = "kupfer_toplevel_" + self._source_config_repr(src)
        self.emit("plugin-toplevel-changed", plugin_id, value)
        return self.set_plugin_config(plugin_id, key,
                                      value, value_type=strbool)

    def get_keybinding(self):
        """Convenience: Kupfer keybinding as string"""
        return self.get_config("Kupfer", "keybinding")

    def set_keybinding(self, keystr):
        """Convenience: Set Kupfer keybinding as string"""
        return self._set_config("Kupfer", "keybinding", keystr)

    def get_magic_keybinding(self):
        """Convenience: Kupfer alternate keybinding as string"""
        return self.get_config("Kupfer", "magickeybinding")

    def set_magic_keybinding(self, keystr):
        """Convenience: Set alternate keybinding as string"""
        return self._set_config("Kupfer", "magickeybinding", keystr)

    def get_global_keybinding(self, key):
        M = {
            "keybinding": self.get_keybinding,
            "magickeybinding": self.get_magic_keybinding,
        }
        return M[key]()

    def set_global_keybinding(self, key, val):
        M = {
            "keybinding": self.set_keybinding,
            "magickeybinding": self.set_magic_keybinding,
        }
        return M[key](val)

    def get_use_command_keys(self):
        return self.get_config("Kupfer", "usecommandkeys")

    def set_use_command_keys(self, enabled):
        return self._set_config("Kupfer", "usecommandkeys", enabled)

    def get_action_accelerator_modifer(self):
        return self.get_config("Kupfer", "action_accelerator_modifer")

    def set_action_accelerator_modifier(self, value):
        """
        Valid values are:
        'alt', 'ctrl'
        """
        return self._set_config("Kupfer", "action_accelerator_modifer", value)

    def set_large_icon_size(self, size):
        return self._set_config("Appearance", "icon_large_size", size)

    def set_small_icon_size(self, size):
        return self._set_config("Appearance", "icon_small_size", size)

    def get_show_status_icon(self):
        """Convenience: Show icon in notification area as bool
        (GTK)
        """
        return strbool(self.get_config("Kupfer", "showstatusicon"))

    def set_show_status_icon(self, enabled):
        """Set config value and return success"""
        return self._set_config("Kupfer", "showstatusicon", enabled)

    def get_show_status_icon_ai(self):
        """Convenience: Show icon in notification area as bool
        (AppIndicator3)
        """
        return strbool(self.get_config("Kupfer", "showstatusicon_ai"))

    def set_show_status_icon_ai(self, enabled):
        """Set config value and return success"""
        return self._set_config("Kupfer", "showstatusicon_ai", enabled)

    def get_directories(self, direct=True):
        """Yield directories to use as directory sources"""

        specialdirs = dict((k, getattr(GLib.UserDirectory, k))
                for k in dir(GLib.UserDirectory) if k.startswith("DIRECTORY_"))

        def get_special_dir(opt):
            if opt.startswith("USER_"):
                _, opt = opt.split("USER_", 1)
                if opt in specialdirs:
                    return GLib.get_user_special_dir(specialdirs[opt])

        level = "Direct" if direct else "Catalog"
        for direc in self.get_config("Directories", level):
            dpath = get_special_dir(direc)
            yield dpath or os.path.abspath(os.path.expanduser(direc))

    def set_directories(self, dirs):
        return self._set_config("Directories", "direct", dirs)

    def get_plugin_config(self, plugin, key, value_type=str, default=None):
        """Return setting @key for plugin names @plugin, try
        to coerce to type @value_type.
        Else return @default if does not exist, or can't be coerced
        """
        plug_section = "plugin_%s" % plugin
        if not plug_section in self._config:
            return default
        val = self._get_raw_config(plug_section, key)

        if val is None:
            return default

        if hasattr(value_type, "load"):
            val_obj = value_type()
            val_obj.load(plugin, key, val)
            return val_obj
        else:
            if value_type is bool:
                value_type = strbool

            try:
                val = value_type(val)
            except ValueError as err:
                self.output_info("Error for stored value %s.%s" %
                        (plug_section, key), err)
                return default
            return val

    def set_plugin_config(self, plugin, key, value, value_type=str):
        """Try set @key for plugin names @plugin, coerce to @value_type
        first.  """
        plug_section = "plugin_%s" % plugin
        self._emit_value_changed(plug_section, key, value)

        if hasattr(value_type, "save"):
            value_repr = value.save(plugin, key)
        else:
            value_repr = value_type(value)
        return self._set_raw_config(plug_section, key, value_repr)

    def get_accelerator(self, name):
        return self.get_config("Keybindings", name)

    def set_accelerator(self, name, key):
        return self._set_config("Keybindings", name, key)

    def get_accelerators(self):
        return self._config['Keybindings']

    def reset_keybindings(self):
        self.set_keybinding(self.get_from_defaults('Kupfer', 'keybinding'))
        self.set_magic_keybinding(
            self.get_from_defaults('Kupfer', 'magickeybinding'))

    def reset_accelerators(self):
        for key, value in self.get_from_defaults('Keybindings'):
            self._set_config('Keybindings', key, value)

    def get_preferred_tool(self, tool_id):
        """
        Get preferred ID for a @tool_id

        Supported: 'terminal'
        """
        return self.get_config("Tools", tool_id)

    def set_preferred_tool(self, tool_id, value):
        return self._set_config("Tools", tool_id, value)

    ## Alternatives section
    ## Provide alternatives for each category
    ## for example the category "terminal"
    def get_valid_alternative_ids(self, category_key):
        """
        Get a list of (id_, name) tuples for the given @category_key
        """
        if not category_key in self._alternative_validators:
            return
        validator = self._alternative_validators[category_key]
        for (id_, alternative) in self._alternatives[category_key].items():
            name = alternative["name"]
            if not validator or validator(alternative):
                yield (id_, name)

    def get_all_alternatives(self, category_key):
        return self._alternatives[category_key]

    def get_preferred_alternative(self, category_key):
        """
        Get preferred alternative dict for @category_key
        """
        tool_id = self.get_preferred_tool(category_key)
        alternatives = self._alternatives[category_key]
        alt = alternatives.get(tool_id)
        if not alt:
            self.output_debug("Warning, no configuration for", category_key)
        return alt or next(iter(alternatives.values()), None)

    def _update_alternatives(self, category_key, alternatives, validator):
        self._alternatives[category_key] = alternatives
        self._alternative_validators[category_key] = validator
        self.emit("alternatives-changed::"+category_key, category_key)


# Arguments: Section, Key, Value
# Detailed by 'section.key' in lowercase
GObject.signal_new("value-changed", SettingsController,
        GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
        GObject.TYPE_BOOLEAN, (GObject.TYPE_STRING, GObject.TYPE_STRING,
        GObject.TYPE_PYOBJECT))

# Arguments: Plugin ID, Value
GObject.signal_new("plugin-enabled-changed", SettingsController,
        GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_INT))

# Arguments: Plugin ID, Value
GObject.signal_new("plugin-toplevel-changed", SettingsController,
        GObject.SignalFlags.RUN_LAST, GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_INT))

# Arguments: Alternative-category
# Detailed by: category key, in lowercase
GObject.signal_new("alternatives-changed", SettingsController,
        GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, ))

_settings_controller = None
def GetSettingsController():
    global _settings_controller
    if _settings_controller is None:
        _settings_controller = SettingsController()
    return _settings_controller



class ExtendedSetting(object):
    """ Abstract class for defining non-simple configuration option """
    def load(self, plugin_id, key, config_value):
        ''' load value for @plugin_id and @key, @config_value is value
        stored in regular Kupfer config for plugin/key'''
        pass

    def save(self, plugin_id, key):
        ''' Save value for @plugin_id and @key.
        @Return value that should be stored in Kupfer config for
        plugin/key (string)'''
        return None




__kupfer_name__ = _("Nemo")
__kupfer_sources__ = ("FmObjects", )
__kupfer_actions__ = (
    "Reveal",
    "GetInfo",
    "CopyTo",
)
__description__ = _("File manager actions")
__version__ = "2017.1"
__author__ = ""

import os

import dbus

from kupfer.objects import Action, Source
from kupfer.objects import FileLeaf, RunnableLeaf, OperationError
from kupfer.obj.apps import AppLeafContentMixin
from kupfer import plugin_support
from kupfer import pretty

plugin_support.check_dbus_connection()

SERVICE_NAME = "org.Nemo"
FO_OBJECT = "/org/Nemo"
FO_IFACE = "org.Nemo.FileOperations"

FM_OBJECT = "/org/freedesktop/FileManager1"
FM_IFACE = "org.freedesktop.FileManager1"

def _get_fm1():
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(SERVICE_NAME, FM_OBJECT)
    except dbus.DBusException as exc:
        raise OperationError(exc)
    iface_obj = dbus.Interface(proxy_obj, FM_IFACE)
    return iface_obj

def _get_nemo():
    bus = dbus.SessionBus()
    try:
        proxy_obj = bus.get_object(SERVICE_NAME, FO_OBJECT)
    except dbus.DBusException as exc:
        raise OperationError(exc)
    iface_obj = dbus.Interface(proxy_obj, FO_IFACE)
    return iface_obj

def _dummy(*args):
    pass

def make_error_handler(ctx):
    def error_handler(exc):
        ctx.register_late_error(exc)
    return error_handler

class Reveal (Action):
    def __init__(self):
        Action.__init__(self, _("Select in File Manager"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        return self.activate_multiple((leaf, ), ctx)

    def activate_multiple(self, leaves, ctx):
        uris = [leaf_uri(leaf) for leaf in leaves]
        id_ = ctx.environment.get_startup_notification_id()
        _get_fm1().ShowItems(uris, id_,
                             reply_handler=_dummy,
                             error_handler=make_error_handler(ctx))

    def item_types(self):
        yield FileLeaf

class GetInfo (Action):
    def __init__(self):
        Action.__init__(self, _("Show Properties"))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        return self.activate_multiple((leaf, ), ctx)

    def activate_multiple(self, leaves, ctx):
        uris = [leaf_uri(leaf) for leaf in leaves]
        id_ = ctx.environment.get_startup_notification_id()
        _get_fm1().ShowItemProperties(uris, id_,
                                      reply_handler=_dummy,
                                      error_handler=make_error_handler(ctx))

    def item_types(self):
        yield FileLeaf

    def get_description(self):
        return _("Show information about file in file manager")

    def get_icon_name(self):
        return "dialog-information"


def _good_destination(dpath, spath):
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to. 
    """
    if not os.path.isdir(dpath):
        return False
    spath = os.path.normpath(spath)
    dpath = os.path.normpath(dpath)
    cpfx = os.path.commonprefix((spath, dpath))
    if os.path.samefile(dpath, spath) or cpfx == spath:
        return False
    return True

def leaf_uri(leaf):
    return leaf.get_gfile().get_uri()

class CopyTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        return self.activate_multiple((leaf, ), (iobj, ), ctx)

    def activate_multiple(self, leaves, iobjects, ctx):
        # Unroll by looping over the destinations,
        # copying everything into each destination
        fm = _get_nemo()
        source_uris = [leaf_uri(L) for L in leaves]

        def _reply(*args):
            self.output_debug("reply got for copying", *args)

        for dest_iobj in iobjects:
            desturi = leaf_uri(dest_iobj)
            fm.CopyURIs(source_uris, desturi,
                        reply_handler=_reply,
                        error_handler=make_error_handler(ctx))

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return True
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Copy file to a chosen location")

class EmptyTrash (RunnableLeaf):
    def __init__(self):
        RunnableLeaf.__init__(self, None, _("Empty Trash"))

    def wants_context(self):
        return True

    def run(self, ctx):
        _get_nemo().EmptyTrash(reply_handler=_dummy,
                               error_handler=make_error_handler(ctx))

    def get_description(self):
        return None

    def get_icon_name(self):
        return "user-trash-full"

class FmObjects(AppLeafContentMixin, Source):
    appleaf_content_id = "nemo"
    def __init__(self):
        Source.__init__(self, _("Nemo"))

    def get_items(self):
        yield EmptyTrash()

    def provides(self):
        yield RunnableLeaf

    def get_icon_name(self):
        return "nemo"





import urllib.parse
from urllib.parse import urlparse as _urlparse
from urllib.parse import urlunparse as _urlunparse

from kupfer import pretty

QFURL_SCHEME = "qpfer"

# One would hope that there was a better way to do this
urllib.parse.uses_netloc.append(QFURL_SCHEME)
try:
    urllib.parse.uses_fragment.append(QFURL_SCHEME)
except AttributeError:
    # Python 2.7.3 drops `uses_fragment` global
    pass

class QfurlError (Exception):
    pass

class qfurl (object):
    """A qfurl is a URI to locate unique objects in kupfer's catalog.

    The qfurl is built up as follows:
    ``qpfer://mother/qfid#module_and_type_hint``

    The mother part is a mother source identifier and is optional.
    The module_and_type_hint is optional.

    A short url looks like the following:
    ``qpfer:identifier``

    This class provides methods to get the qfurl for an object,
    and resolve the object in a catalog.

    >>> class Object (object):
    ...     qf_id = "token"
    ...
    >>> q = qfurl(Object())
    >>> qfurl.reduce_url(q.url)
    'qpfer:token'

    >>> class Source (object):
    ...     def get_leaves(self):
    ...         yield Object()
    ...     def provides(self):
    ...         yield Object
    ...
    >>> q.resolve_in_catalog((Source(), ))  # doctest: +ELLIPSIS
    <__main__.Object object at 0x...>
    """

    def __init__(self, obj=None, url=None):
        """Create a new qfurl for object @obj"""
        if obj:
            typname = "%s.%s" % (type(obj).__module__, type(obj).__name__)
            try:
                qfid = obj.qf_id
            except AttributeError:
                raise QfurlError("%s has no qfurl" % obj)
            self.url = _urlunparse((QFURL_SCHEME, "", qfid, "", "", typname))
        else:
            self.url = url

    def __str__(self):
        return self.url

    def __hash__(self):
        return hash(self.url)

    def __eq__(self, other):
        return self.reduce_url(self.url) == self.reduce_url(other.url)

    @classmethod
    def reduce_url(cls, url):
        """
        >>> url = "qpfer://mother/qfid#module_and_type_hint"
        >>> qfurl.reduce_url(url)
        'qpfer://mother/qfid'
        """
        return urllib.parse.urldefrag(url)[0].replace("///", "", 1)

    @classmethod
    def _parts_mother_id_typename(cls, url):
        """
        >>> murl = "qpfer://mother/qfid#module_and_type_hint"
        >>> qfurl._parts_mother_id_typename(murl)
        ('mother', 'qfid', 'module_and_type_hint')
        """
        scheme, mother, qfid, _ign, _ore, typname = _urlparse(url)
        if scheme != QFURL_SCHEME:
            raise QfurlError("Wrong scheme: %s" % scheme)
        qfid = qfid.lstrip("/")
        return mother, qfid, typname

    def resolve_in_catalog(self, catalog):
        """Resolve self in a catalog of sources

        Return *immediately* on match found"""
        mother, qfid, typname = self._parts_mother_id_typename(self.url)
        module, name = typname.rsplit(".", 1) if typname else (None, None)
        for src in catalog:
            if name:
                if name not in (pt.__name__
                        for pt in src.provides()) and \
                    name not in (t.__name__
                        for pt in src.provides()
                        for t in pt.__subclasses__()):
                    continue
            for obj in src.get_leaves():
                if not hasattr(obj, "qf_id"):
                    continue
                try:
                    if self == qfurl(obj):
                        return obj
                except QfurlError:
                    pass
        pretty.print_debug(__name__, "No match found for", self)
        return None

if __name__ == '__main__':
    import doctest
    doctest.testmod()




from gi.repository import GLib, GObject


from kupfer import pretty
from kupfer.weaklib import gobject_connect_weakly

_scheduler = None

def GetScheduler():
    """Get the shared instance"""
    global _scheduler
    if not _scheduler:
        _scheduler = Scheduler()
    return _scheduler

class Scheduler (GObject.GObject, pretty.OutputMixin):
    __gtype_name__ = "Scheduler"
    def __init__(self):
        super(Scheduler, self).__init__()
        self._finished = False

    def load(self):
        self.output_debug("Loading")
        self.emit("load")
        self.emit("loaded")
        self.output_debug("Loaded")

    def display(self):
        self.output_debug("Display")
        self.emit("display")
        GLib.idle_add(self._after_display)

    def _after_display(self):
        self.output_debug("After Display")
        self.emit("after-display")

    @property
    def finished(self):
        return self._finished

    def finish(self):
        self._finished = True
        self.emit("finish")
GObject.signal_new("load", Scheduler, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())
GObject.signal_new("loaded", Scheduler, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())
GObject.signal_new("display", Scheduler, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())
GObject.signal_new("after-display", Scheduler, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())
GObject.signal_new("finish", Scheduler, GObject.SignalFlags.RUN_LAST,
        GObject.TYPE_BOOLEAN, ())

class Timer (object):
    def __init__(self, call_at_finish=False):
        self._current_timer = -1
        self._call_at_finish = call_at_finish
        self._current_callback = None
        gobject_connect_weakly(GetScheduler(), "finish", self._on_finish)

    def set(self, timeout_seconds, callback, *arguments):
        """Setup timer to call @timeout_seconds in the future.
        If the timer was previously set, it is postponed
        """
        self.invalidate()
        self._current_callback = lambda : callback(*arguments)
        self._current_timer = GLib.timeout_add_seconds(timeout_seconds,
                self._call)

    def set_ms(self, timeout_milliseconds, callback, *arguments):
        """Setup timer to call @timeout_milliseconds in the future.
        If the timer was previously set, it is postponed
        """
        self.invalidate()
        self._current_callback = lambda : callback(*arguments)
        self._current_timer = GLib.timeout_add(int(timeout_milliseconds),
                self._call)

    def set_idle(self, callback, *arguments):
        self.invalidate()
        self._current_callback = lambda : callback(*arguments)
        self._current_timer = GLib.idle_add(self._call)

    def _call(self, timer=None):
        self._current_timer = -1
        self._current_callback()
    
    def invalidate(self):
        if self._current_timer > 0:
            GLib.source_remove(self._current_timer)
        self._current_timer = -1

    def is_valid(self):
        """If Timer is currently set"""
        return (self._current_timer > 0)

    def _on_finish(self, scheduler):
        if self._call_at_finish and self.is_valid():
            self._call()
        else:
            self.invalidate()



"""
This is a program of its own, that does not integrate with the
Kupfer process.
"""
import builtins
import os

import gi

gi.require_version("Gtk", "3.0")
gi.require_version("Keybinder", "3.0")

from gi.repository import Gtk, Keybinder as keybinder

import dbus

from dbus.mainloop.glib import DBusGMainLoop

SERV = "se.kaizer.kupfer"
OBJ = "/interface"
IFACE = "se.kaizer.kupfer.Listener"

if not hasattr(builtins, '_'):
    def _(x):
        return x

def get_all_keys():
    try:
        bus = dbus.Bus()
        obj = bus.get_object(SERV, OBJ)
        iface = dbus.Interface(obj, IFACE)
        return iface.GetBoundKeys(byte_arrays=True)
    except dbus.DBusException as exc:
        print(exc)
        print("Waiting for Kupfer to start..")
        return []

def rebind_key(keystring, is_bound):
    if is_bound:
        print("binding", keystring)
        keybinder.bind(keystring, relay_key, keystring)
    else:
        print("unbinding", keystring)
        keybinder.unbind(keystring)

def relay_key(key):
    print("Relaying", key)
    time = keybinder.get_current_event_time()
    s_id = "kupfer-%d_TIME%s" % (os.getpid(), time)
    bus = dbus.Bus()
    obj = bus.get_object(SERV, OBJ, introspect=False)
    iface = dbus.Interface(obj, IFACE)
    iface.RelayKeysFromDisplay(key, os.getenv("DISPLAY", ":0"), s_id)

def main():
    DBusGMainLoop(set_as_default=True)

    relayed_keys = list(get_all_keys())

    for key in relayed_keys:
        rebind_key(key, True)
    bus = dbus.Bus()
    bus.add_signal_receiver(rebind_key, 'BoundKeyChanged',
            dbus_interface=IFACE)
    sicon = Gtk.StatusIcon.new_from_icon_name("kupfer")
    display = os.getenv("DISPLAY", ":0")
    sicon.set_tooltip_text(_("Keyboard relay is active for display %s") % display)
    sicon.set_visible(True)
    try:
        Gtk.main()
    except KeyboardInterrupt:
        raise SystemExit(0)

if __name__ == '__main__':
    main()




def _extend_path():
    # Inside a function to not leak variables to module namespace
    import os
    from kupfer import config
    from kupfer import pretty

    if not config.has_capability("CUSTOM_PLUGINS"):
        return

    # Add plugins in data directories
    __path__.extend(config.get_data_dirs("plugins"))

    # Add .zip files in plugins directories
    for directory in list(__path__):
        try:
            filenames = os.listdir(directory)
        except OSError as error:
            pretty.print_error(__name__, error)
            continue
        zipnames = [f for f in filenames if f.endswith(".zip")]
        if zipnames:
            pretty.print_debug(__name__, "Adding", directory, zipnames)
        __path__.extend(os.path.join(directory, z) for z in zipnames)

_extend_path()




# -*- coding: utf-8 -*

__kupfer_name__ = _("XFCE Session Management")
__kupfer_sources__ = ("XfceItemsSource", )
__description__ = _("Special items and actions for XFCE environment")
__version__ = "2021-04-11"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

from kupfer.plugin import session_support as support
from kupfer import plugin_support

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "lock_cmd",
        "label": _("Screen lock command"),
        "type": str,
        "value": "xflock4",
    },
)

# sequences of argument lists
LOGOUT_CMD = (["xfce4-session-logout", "--logout"],)
SHUTDOWN_CMD = (["xfce4-session-logout"],)


class XfceItemsSource (support.CommonSource):
    def __init__(self):
        support.CommonSource.__init__(self, _("XFCE Session Management"))

    def get_items(self):
        lockscreen_cmd = __kupfer_settings__['lock_cmd'] \
            or "xdg-screensaver lock"

        return (
            support.Logout(LOGOUT_CMD),
            support.LockScreen((lockscreen_cmd.split(" "), )),
            support.Shutdown(SHUTDOWN_CMD),
        )



# -*- coding: UTF-8 -*-
__kupfer_name__ = _("Shorten Links")
__kupfer_actions__ = ("ShortenLinks", )
__description__ = _("Create short aliases of long URLs")
__version__ = "2017.1"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

import urllib.request, urllib.parse

from kupfer.objects import Leaf, Action, Source, UrlLeaf, OperationError
from kupfer import pretty

class _ShortLinksService(Leaf):
    def __init__(self, name):
        Leaf.__init__(self, name, name)
    def get_icon_name(self):
        return "text-html"

class _GETService(_ShortLinksService, pretty.OutputMixin):
    """ A unified shortener service working with GET requests """
    host = None
    path = None
    url_key = "url"
    use_https = False

    def process(self, url):
        """Shorten @url or raise ValueError"""
        query_string = urllib.parse.urlencode({self.url_key : url})
        try:
            pretty.print_debug(__name__, "Request", self.path + query_string)
            resp = urllib.request.urlopen(self.path + query_string)
            if resp.status != 200:
                raise ValueError('Invalid response %d, %s' % (resp.status, resp.reason))
            
            result = resp.read()
            return result.strip().decode("utf-8")

        except (OSError, IOError, ValueError) as exc:
            raise ValueError(exc)
        return _('Error')


# NOTE: It's important that we use only sites that provide a stable API

class IsGd(_GETService):
    """
    Website: http://is.gd
    Reference: http://is.gd/apishorteningreference.php
    """
    path = 'https://is.gd/create.php?format=simple&'

    def __init__(self):
        _ShortLinksService.__init__(self, 'Is.gd')

class VGd(_GETService):
    """
    Website: http://v.gd
    Reference: http://v.gd/apishorteningreference.php

    Like is.gd, but v.gd always shows a preview page.
    """
    path = 'https://v.gd/create.php?format=simple&'

    def __init__(self):
        _ShortLinksService.__init__(self, 'V.gd')


class ShortenLinks(Action):
    ''' Shorten links with selected engine '''

    def __init__(self):
        Action.__init__(self, _('Shorten With...'))

    def has_result(self):
        return True

    def activate(self, leaf, iobj):
        try:
            result = iobj.process(leaf.object)
        except ValueError as exc:
            raise OperationError(str(exc))
        return UrlLeaf(result, result)

    def item_types(self):
        yield UrlLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield _ShortLinksService

    def object_source(self, for_item=None):
        return ServicesSource()

    def get_description(self):
        return __description__


class ServicesSource(Source):
    source_use_cache = False
    def __init__(self):
        super().__init__(_("Services"))

    def get_items(self):
        yield IsGd()
        yield VGd()

    def should_sort_lexically(self):
        return True

    def get_icon_name(self):
        return "applications-internet"



# -*- coding: UTF-8 -*-


__kupfer_name__ = _("Top")
__kupfer_sources__ = ("TaskSource", )
__description__ = _("Show running tasks and allow sending signals to them")
__version__ = "2009-11-24"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os
import signal
import operator

from kupfer.objects import Action, Source, Leaf
from kupfer import scheduler
from kupfer import plugin_support
from kupfer import utils

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "sort_order",
        "label": _("Sort Order"),
        "type": str,
        "value": _("Commandline"),
        "alternatives": [_("Commandline"), _("CPU usage (descending)"),
                _("Memory usage (descending)") ]
    },
)


class Task(Leaf):
    def __init__(self, path, name, description=None):
        Leaf.__init__(self, path, name)
        self._description = description

    def get_description(self):
        return self._description

    def get_actions(self):
        yield SendSignal()

    def get_icon_name(self):
        return 'applications-system'


class SendSignal(Action):
    def __init__(self):
        Action.__init__(self, _("Send Signal..."))

    def activate(self, leaf, iobj):
        os.kill(leaf.object, iobj.object)

    def requires_object(self):
        return True
    
    def object_types(self):
        yield _Signal

    def object_source(self, for_item=None):
        return _SignalsSource()


class _Signal(Leaf):
    def get_description(self):
        return "kill -%s ..." % self.object


# get all signals from signal package
_SIGNALS = [
    _Signal(getattr(signal, signame), signame[3:])
    for signame in sorted(dir(signal))
    if signame.startswith('SIG') and not signame.startswith('SIG_')
]


class _SignalsSource(Source):
    def __init__(self):
        Source.__init__(self, _("Signals"))

    def get_items(self):
        return _SIGNALS

    def provides(self):
        yield _Signal
    

class TaskSource(Source):
    task_update_interval_sec = 5
    source_use_cache = False

    def __init__(self, name=_("Running Tasks")):
        Source.__init__(self, name)
        self._cache = []
        self._version = 2

    def initialize(self):
        self._timer = scheduler.Timer()

    def finalize(self):
        self._timer = None
        self._cache = []

    def _async_top_finished(self, acommand, stdout, stderr):
        self._cache = []

        processes = parse_top_output(stdout)
        # sort processes (top don't allow to sort via cmd line)
        if __kupfer_settings__['sort_order'] == _("Memory usage (descending)"):
            processes = sorted(processes, key=operator.itemgetter(2),
                               reverse=True)
        elif __kupfer_settings__['sort_order'] == _("Commandline"):
            processes = sorted(processes, key=operator.itemgetter(4))
        # default: by cpu

        fields = _("pid: %(pid)s  cpu: %(cpu)g%%  mem: %(mem)g%%  time: %(time)s")
        for pid, cpu, mem, ptime, cmd in processes:
            description = fields % dict(pid=pid, cpu=cpu, mem=mem, time=ptime)
            self._cache.append(Task(pid, cmd, description))

        self.mark_for_update()

    def _async_top_start(self):
        uid = os.getuid()
        utils.AsyncCommand(["top", "-b", "-n", "1", "-u", "%d" % uid],
                           self._async_top_finished, 60, env=["LC_NUMERIC=C"])

    def get_items(self):
        for task in self._cache:
            yield task
        update_wait = self.task_update_interval_sec if self._cache else 0
        # update after a few seconds
        self._timer.set(update_wait, self._async_top_start)

    def get_description(self):
        return _("Running tasks for current user")

    def get_icon_name(self):
        return "system"

    def provides(self):
        yield Task


def parse_top_output(out):
    """
    Yield tuples of (pid, cpu, mem, ptime, cmd)
    """
    # Assuming UTF-8 output
    fields_map = None
    fields_count = 0
    header_read = False
    for line in out.split(b'\n'):
        line = line.decode("utf-8", "replace").strip()
        if line == '':
            header_read = True
            continue

        if not header_read:
            continue

        if line.startswith('PID'): # assume pid is first col
            fields_map = dict(((name, pos) for pos, name in enumerate(line.split())))
            fields_count = len(fields_map)
            continue    # skip header

        line_fields = line.split(None, fields_count-1)
        pid = line_fields[0]
        cpu = line_fields[fields_map['%CPU']]
        mem = line_fields[fields_map['%MEM']]
        ptime = line_fields[fields_map['TIME+']]
        cmd = line_fields[-1]

        # read command line
        proc_file = '/proc/%s/cmdline' % pid
        if os.path.isfile(proc_file): # also skip (finished) missing tasks
            with open(proc_file, 'rt') as f:
                cmd = f.readline().replace('\x00', ' ') or cmd

            yield (int(pid), float(cpu), float(mem), ptime, cmd)




__kupfer_name__ = _("Devhelp")
__kupfer_actions__ = ("LookUp", )
__description__ = _("Search in Devhelp")
__version__ = "2017.1"
__author__ = ""

from kupfer.objects import Action, TextLeaf, OperationError
from kupfer import utils


class LookUp (Action):
    def __init__(self):
        Action.__init__(self, _("Search in Devhelp"))
    def activate(self, leaf):
        text = leaf.object
        try:
            utils.spawn_async_raise(['devhelp', '--search=%s' % text])
        except utils.SpawnError as exc:
            raise OperationError(exc)
    def item_types(self):
        yield TextLeaf
    def valid_for_item(self, leaf):
        text = leaf.object
        return '\n' not in text
    def get_description(self):
        return None
    def get_icon_name(self):
        return "devhelp"



__kupfer_name__ = _("Archive Manager")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
        "UnpackHere",
        "CreateArchive",
        "CreateArchiveIn",
    )
__description__ = _("Use Archive Manager actions")
__version__ = ""
__author__ = "Ulrik"

import os
import re
# since "path" is a very generic name, you often forget..
from os import path as os_path

from kupfer.objects import Action, FileLeaf
from kupfer import utils 
from kupfer import plugin_support
from kupfer import runtimehelper


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "archive_type",
        "label": _("Compressed archive type for 'Create Archive In'"),
        "type": str,
        "value": ".tar.gz",
        "alternatives": (
            ".7z",
            ".rar",
            ".tar",
            ".tar.gz",
            ".tar.bz2",
            ".tar.xz",
            ".zip",
            )
    },
)

class UnpackHere (Action):
    def __init__(self):
        Action.__init__(self, _("Extract Here"))
        self.extensions_set = set((
            ".rar", ".7z", ".zip", ".gz", ".tgz", ".tar", ".lzma", ".bz2",
            ".tbz2", ".tzo", ".lzo", ".xz", ".ar", ".cbz", ".Z", ".taz",
            ".lz", ".bz", ".tbz", ".lzh",
            ))
    def activate(self, leaf):
        utils.spawn_async_notify_as("file-roller.desktop",
                ["file-roller", "--extract-here", leaf.object])

    def valid_for_item(self, item):
        tail, ext = os.path.splitext(item.object)
        # FIXME: Make this detection smarter
        # check for standard extension or a multi-part rar extension
        return (ext.lower() in self.extensions_set or
            re.search(r".r\d+$", ext.lower()) is not None)

    def item_types(self):
        yield FileLeaf
    def get_description(self):
        return _("Extract compressed archive")
    def get_icon_name(self):
        return "extract-archive"

class CreateArchive (Action):
    def __init__(self):
        Action.__init__(self, _("Create Archive"))

    @classmethod
    def _make_archive(cls, filepaths):
        cmd = ["file-roller", "--add"]
        cmd.extend(filepaths)
        utils.spawn_async_notify_as("file-roller.desktop", cmd)

    def activate(self, leaf):
        self._make_archive((leaf.object, ))
    def activate_multiple(self, objs):
        self._make_archive([L.object for L in objs])

    def item_types(self):
        yield FileLeaf
    def get_description(self):
        return _("Create a compressed archive from folder")
    def get_icon_name(self):
        return "add-files-to-archive"

class CreateArchiveIn (Action):
    def __init__(self):
        Action.__init__(self, _("Create Archive In..."))

    @classmethod
    def _make_archive(cls, ctx, basename, dirpath, filepaths):
        archive_type = __kupfer_settings__["archive_type"]
        archive_path = \
            utils.get_destpath_in_directory(dirpath, basename, archive_type)
        cmd = ["file-roller", "--add-to=%s" % (archive_path, )]
        cmd.extend(filepaths)
        runtimehelper.register_async_file_result(ctx, archive_path)
        utils.spawn_async_notify_as("file-roller.desktop", cmd)
        return archive_path

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        dirpath = iobj.object
        basename = os_path.basename(leaf.object)
        self._make_archive(ctx, basename, dirpath, (leaf.object, ))

    def activate_multiple(self, objs, iobjs, ctx):
        # TRANS: Default filename (no extension) for 'Create Archive In...'
        basename = _("Archive")
        for iobj in iobjs:
            self._make_archive(ctx, basename, iobj.object,
                               [L.object for L in objs])

    def item_types(self):
        yield FileLeaf
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item=None):
        return utils.is_directory_writable(obj.object)
    def get_description(self):
        return _("Create a compressed archive from folder")
    def get_icon_name(self):
        return "add-files-to-archive"




# -*- coding: UTF-8 -*-
'''
virtualbox_vboxapi_support.py

Control VirtualBox via Python interface (vboxapi).
Only (?) Sun VirtualBox (no OSE).
'''
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"
__version__ = "2018-09-07"


from kupfer import pretty

import vboxapi


from kupfer.plugin.virtualbox import constants as vbox_const

MONITORED_DIRS = None
IS_DYNAMIC = False
ICON = "VBox"
APP_ID = "virtualbox"

_ACTIONS = {
        vbox_const.VM_POWEROFF: lambda c: c.powerDown(),
        vbox_const.VM_ACPI_POWEROFF: lambda c: c.powerButton(),
        vbox_const.VM_PAUSE: lambda c: c.pause(),
        vbox_const.VM_REBOOT: lambda c: c.reset(),
        vbox_const.VM_RESUME: lambda c: c.resume(),
        vbox_const.VM_SAVE: lambda c: c.saveState(),
}


def _get_object_session():
    ''' get new session to vm '''
    pretty.print_debug(__name__, '_get_object_session start')
    vbox, session = None, None
    try:
        vbox = vboxapi.VirtualBoxManager(None, None)
        session = vbox.mgr.getSessionObject(vbox.vbox)
    except Exception as err:
        pretty.print_error(__name__, 'virtualbox: get session error ', err)
    pretty.print_debug(__name__, '_get_object_session finished', vbox, session)
    return vbox, session


def _get_existing_session(vm_uuid):
    ''' get existing session by machine uuid '''
    pretty.print_debug(__name__, '_get_existing_session start')
    vbox, session = None, None
    try:
        vbox = vboxapi.VirtualBoxManager(None, None)
        session = vbox.mgr.getSessionObject(vbox.vbox)
        vbox.vbox.openExistingSession(session, vm_uuid)
    except Exception as err:
        pretty.print_error(__name__, 'virtualbox: get session error', vm_uuid,
                err)
    pretty.print_debug(__name__, '_get_existing_session finished', vbox, session)
    return vbox, session


def get_machine_state(machine_id):
    ''' check vms state (on/off/paused) '''
    pretty.print_debug(__name__, 'get_machine_state', machine_id)
    vbox, vbox_sess = _get_object_session()
    if vbox_sess is None:
        return vbox_const.VM_STATE_POWEROFF
    state = vbox_const.VM_STATE_POWERON
    try:
        vbox.vbox.openExistingSession(vbox_sess, machine_id)
        machine_state = vbox_sess.machine.state
        if machine_state == vbox.constants.MachineState_Paused:
            state = vbox_const.VM_STATE_PAUSED
        elif machine_state in (vbox.constants.MachineState_PoweredOff,
                vbox.constants.MachineState_Aborted,
                vbox.constants.MachineState_Starting):
            state = vbox_const.VM_STATE_POWEROFF
        elif machine_state == vbox.constants.MachineState_Saved:
            state = vbox_const.VM_STATE_SAVED
    except Exception as err:  # exception == machine is off (xpcom.Exception)
        # silently set state to off
        state = vbox_const.VM_STATE_POWEROFF
        pretty.print_debug(__name__, 'get_machine_state error', err)
    try:
        if vbox_sess.state == vbox.constants.SessionState_Open:
            vbox_sess.close()
    except Exception:  # varoius errors (xpcom.Exception)
        pass
    pretty.print_debug(__name__, 'get_machine_state finish', machine_id, state)
    return state


def _machine_start(vm_uuid, mode):
    ''' Start virtual machine
        @param vm_uuid - uuid of virtual machine
        @param mode - mode: gui, headless
    '''
    vbox, session = _get_object_session()
    if session:
        try:
            remote_sess = vbox.vbox.openRemoteSession(session, vm_uuid, mode,
                    '')
            remote_sess.waitForCompletion(-1)
        except Exception as err:
            pretty.print_error(__name__, "StartVM:", vm_uuid, "Mode ", mode,
                    "error", err)
        try:
            if session.state == vbox.constants.SessionState_Open:
                session.close()
        except Exception:  # varoius errors (xpcom.Exception)
            pass


def _execute_machine_action(vm_uuid, action):
    ''' Start virtual machine
        @param vm_uuid - uuid of virtual machine
        @param action - function called on vbox session
    '''
    vbox, session = _get_existing_session(vm_uuid)
    try:
        action(session.console)
    except Exception as err:
        pretty.print_error(__name__, "_execute_machine_action:", repr(action),
                " vm:", vm_uuid, "error", err)
    try:
        if session.state == vbox.constants.SessionState_Open:
            session.close()
    except Exception:  # varoius errors (xpcom.Exception)
        pass


def vm_action(action, vm_uuid):
    ''' change state of the virtual machine
        @param action - one of the const VM_*
        @param vm_uuid - virtual machine uuid
    '''
    if action == vbox_const.VM_START_NORMAL:
        _machine_start(vm_uuid, 'gui')
    elif action == vbox_const.VM_START_HEADLESS:
        _machine_start(vm_uuid, 'headless')
    else:
        command = _ACTIONS[action]
        _execute_machine_action(vm_uuid, command)


def get_machines():
    ''' Get generator of items:
        (machine uuid, machine name, machine description)
    '''
    pretty.print_debug(__name__, 'get_machines start')
    vbox, vbox_sess = _get_object_session()
    if vbox_sess is None:
        return
    machines = vbox.getArray(vbox.vbox, 'machines')
    for machine in machines:
        pretty.print_debug(__name__, 'get_machines; found machine',
                machine.id, machine.name)
        description = machine.description or machine.OSTypeId
        yield (machine.id, machine.name, description)
    pretty.print_debug(__name__, 'get_machines finished')


def unload():
    pass





debug = False

import sys
import traceback
from time import time as timestamp

class OutputMixin (object):
    """
    A mixin class providing prefixed output
    standard output and debug output
    """
    def _output_category(self):
        return "[%s] %s:" % (type(self).__module__, type(self).__name__)

    def _output_core(self, prefix, sep, end, stream, *items):
        category = self._output_category()
        print(prefix+category, *items, sep=sep, end=end, file=stream)

    def output_info(self, *items, **kwargs):
        """
        Output given items using @sep as separator,
        ending the line with @end
        """
        sep = kwargs.get("sep", " ")
        end = kwargs.get("end", "\n")
        self._output_core("", sep, end, sys.stdout, *items)

    def output_exc(self, exc_info=None):
        """Output current exception, or use @exc_info if given"""
        etype, value, tb = (exc_info or sys.exc_info())
        if debug:
            self._output_core("Exception in ", "", "\n", sys.stderr)
            traceback.print_exception(etype, value, tb, file=sys.stderr)
        else:
            msg = "%s: %s" % (etype.__name__, value)
            self._output_core("Exception in ", " ", "\n", sys.stderr, msg)

    def output_debug(self, *items, **kwargs):
        if debug:
            sep = kwargs.get("sep", " ")
            end = kwargs.get("end", "\n")
            self._output_core("D ", sep, end, sys.stderr, *items)

    def output_error(self, *items, **kwargs):
        sep = kwargs.get("sep", " ")
        end = kwargs.get("end", "\n")
        self._output_core("Error ", sep, end, sys.stderr, *items)

class _StaticOutput (OutputMixin):
    current_calling_module = None
    def _output_category(self):
        return "[%s]:" % (self.current_calling_module, )

    def print_info(self, modulename, *args, **kwargs):
        self.current_calling_module = modulename
        self.output_info(*args, **kwargs)

    def print_error(self, modulename, *args, **kwargs):
        self.current_calling_module = modulename
        self.output_error(*args, **kwargs)

    def print_exc(self, modulename, *args, **kwargs):
        self.current_calling_module = modulename
        self.output_exc(*args, **kwargs)

    def print_debug(self, modulename, *args, **kwargs):
        if debug:
            self.current_calling_module = modulename
            self.output_debug(*args, **kwargs)
_StaticOutput = _StaticOutput()

print_info = _StaticOutput.print_info
print_debug = _StaticOutput.print_debug
print_error = _StaticOutput.print_error
print_exc = _StaticOutput.print_exc


def timing_start():
    if debug:
        return [timestamp()]
    else:
        return None

def timing_step(modulename, start, label):
    if debug:
        t = timestamp()
        print_debug(modulename, label, "in %.6f s" % (t - start[0], ))
        start[0] = t



from kupfer import kupferstring
from kupfer.obj.base import OperationError

class LocaleOperationError (OperationError):
    """
    User-visible error created from locale-encoded
    error message (for example OSError)
    """
    def __init__(self, s):
        OperationError.__init__(self, kupferstring.fromlocale(s))

class NotAvailableError (OperationError):
    """
    User-visible error message when an external
    tool is the wrong version
    """
    def __init__(self, toolname):
        OperationError.__init__(self,
                       _("%s does not support this operation") % toolname)

class NoMultiError (OperationError):
    def __init__(self):
        OperationError.__init__(self,
                       _("Can not be used with multiple objects"))



"""
Work with Textfiles: Allow appending and writing new files,
or extracting the content of files.

All Text in Kupfer is in unicode. When we read from textfiles or write
to textfiles, we always work in the locale-defined encoding.

FIXME: Be less strict (use UTF-8 if locale says Ascii)
"""



__kupfer_name__ = _("Textfiles")
__kupfer_actions__ = (
        "AppendTo",
        "AppendText",
        "WriteTo",
        "GetTextContents",
    )
__description__ = None
__version__ = "2017.1"
__author__ = ""

from gi.repository import Gio

from kupfer.objects import Action
from kupfer.objects import TextLeaf, FileLeaf
from kupfer.obj import helplib
from kupfer import kupferstring
from kupfer import utils

# FIXME: Sometimes require that the type is *exactly* text/plain?

class AppendTo (Action):
    def __init__(self, name=None):
        if not name:
            name = _("Append To...")
        Action.__init__(self, name)

    def activate(self, leaf, iobj):
        with open(iobj.object, "a") as outfile:
            outfile.write(leaf.object)
            outfile.write("\n")

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, iobj, for_item=None):
        return iobj.is_content_type("text/plain")

    def get_icon_name(self):
        return "list-add"

class AppendText (helplib.reverse_action(AppendTo)):
    def __init__(self):
        Action.__init__(self, _("Append..."))

class WriteTo (Action):
    def __init__(self):
        Action.__init__(self, _("Write To..."))

    def has_result(self):
        return True

    def activate(self, leaf, iobj):
        outfile, outpath = \
                utils.get_destfile_in_directory(iobj.object, _("Empty File"))
        try:
            l_text = kupferstring.tolocale(leaf.object)
            outfile.write(l_text)
            if not l_text.endswith(b"\n"):
                outfile.write(b"\n")
        finally:
            outfile.close()
        return FileLeaf(outpath)

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, iobj, for_item=None):
        return iobj.is_dir()

    def get_description(self):
        return _("Write the text to a new file in specified directory")

    def get_icon_name(self):
        return "document-new"

class GetTextContents (Action):
    def __init__(self):
        Action.__init__(self, _("Get Text Contents"))

    def has_result(self):
        return True

    def activate(self, leaf):
        with open(leaf.object, "rb") as infile:
            l_text = infile.read()
            us_text = kupferstring.fromlocale(l_text)
        return TextLeaf(us_text)

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return item.is_content_type("text/plain")

    def get_icon_name(self):
        return "edit-copy"



# -*- encoding: utf-8 -*-
"""
Kupfer's Contacts API

Main definition and *constructor* classes.

Constructor classes such as EmailContact are used to conveniently construct
contacts with common traits. To *use* contacts, always use ContactLeaf, asking
for specific slots to be filled.
"""
import re

from kupfer import icons
from kupfer.obj.grouping import GroupingLeaf

__author__ = ("Ulrik Sverdrup <ulrik.sverdrup@gmail.com>, "
              "Karol Będkowski <karol.bedkowsk+gh@gmail.com>",
              "Adi Sieker <adi@sieker.info>",
)

EMAIL_KEY = "EMAIL"
NAME_KEY = "NAME"
PHONE_KEY = "PHONE"
ADDRESS_KEY = "ADDRESS"
LABEL_KEY = "LABEL"
JABBER_JID_KEY = "JID"
JABBER_STATUS_KEY = "JABBER_STATUS"
JABBER_RESOURCE_KEY = "JABBER_RESOURCE"
AIM_KEY = "AIM"
GOOGLE_TALK_KEY = "GOOGLE_TALK"
ICQ_KEY = "ICQ"
MSN_KEY = "MSN"
QQ_KEY = "QQ"
SKYPE_KEY = "SKYPE"
YAHOO_KEY = "YAHOO"


class ContactLeaf(GroupingLeaf):
    grouping_slots = ()

    def __init__(self, obj, name, image=None):
        self.image = image
        GroupingLeaf.__init__(self, obj, name)

    def get_icon_name(self):
        return "stock_person"

    def get_text_representation(self):
        return self.get_description()

    def get_thumbnail(self, width, height):
        if self.image:
            return icons.get_pixbuf_from_data(self.image, width, height)
        return GroupingLeaf.get_thumbnail(self, width, height)

## E-mail convenience and constructors

def _get_email_from_url(url):
    ''' convert http://foo@bar.pl -> foo@bar.pl '''
    sep = url.find('://')
    return url[sep + 3:] if sep > -1 else url

# FIXME: Find a more robust (less strict?) approach than regex
_CHECK_EMAIL_RE = re.compile(r"^[a-z0-9\._%-+]+\@[a-z0-9._%-]+\.[a-z]{2,}$")


def is_valid_email(email):
    ''' simple email check '''
    return len(email) > 7 and _CHECK_EMAIL_RE.match(email.lower()) is not None


def email_from_leaf(leaf):
    """
    Return an email address string if @leaf has a valid email address.

    @leaf may also be a TextLeaf or UrlLeaf.
    Return a false value if no valid email is found.
    """
    if isinstance(leaf, ContactLeaf):
        return EMAIL_KEY in leaf and leaf[EMAIL_KEY]
    email = _get_email_from_url(leaf.object)
    return is_valid_email(email) and email


class EmailContact (ContactLeaf):
    grouping_slots = ContactLeaf.grouping_slots + (EMAIL_KEY, )

    def __init__(self, email, name, image=None):
        slots = {EMAIL_KEY: email, NAME_KEY: name}
        ContactLeaf.__init__(self, slots, name, image)

    def repr_key(self):
        return self.object[EMAIL_KEY]

    def get_description(self):
        return self.object[EMAIL_KEY]

    def get_gicon(self):
        return icons.ComposedIconSmall(self.get_icon_name(), "stock_mail")


class IMContact (ContactLeaf):
    grouping_slots = ContactLeaf.grouping_slots + (EMAIL_KEY, )

    def __init__(self, im_id_kind, im_id, name, label=None, other_slots=None,
            image=None):
        self.im_id_kind = im_id_kind
        slots = {im_id_kind: im_id, NAME_KEY: name, LABEL_KEY: label}
        if other_slots:
            slots.update(other_slots)
        ContactLeaf.__init__(self, slots, name, image)
        self.kupfer_add_alias(im_id)

    def repr_key(self):
        return self.object[self.im_id_kind]

    def get_description(self):
        return self.object[LABEL_KEY] or self.object[self.im_id_kind]


class JabberContact (IMContact):
    ''' Minimal class for all Jabber contacts. '''
    grouping_slots = IMContact.grouping_slots + (JABBER_JID_KEY, )

    def __init__(self, jid, name, status=None, resource=None, slots=None,
            image=None):
        IMContact.__init__(self, JABBER_JID_KEY, jid, name or jid,
                other_slots=slots, image=image)
        self._description = _("[%(status)s] %(userid)s/%(service)s") % \
                {
                    "status": status or _("unknown"),
                    "userid": jid,
                    "service": resource or "",
                }

    def get_description(self):
        return self._description


class AIMContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (AIM_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, AIM_KEY, id_, name, _("Aim"), slots, image)


class GoogleTalkContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (GOOGLE_TALK_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, GOOGLE_TALK_KEY, id_, name, _("Google Talk"),
                slots, image)


class ICQContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (ICQ_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, ICQ_KEY, id_, name, _("ICQ"), slots, image)


class MSNContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (MSN_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, MSN_KEY, id_, name, _("MSN"), slots, image)


class QQContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (QQ_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, QQ_KEY, id_, name, _("QQ"), slots, image)


class YahooContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (YAHOO_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, YAHOO_KEY, id_, name, _("Yahoo"), slots, image)


class SkypeContact(IMContact):
    grouping_slots = IMContact.grouping_slots + (SKYPE_KEY, )

    def __init__(self, id_, name, slots=None, image=None):
        IMContact.__init__(self, SKYPE_KEY, id_, name, _("Skype"), slots, image)


class PhoneContact(ContactLeaf):
    grouping_slots = ContactLeaf.grouping_slots + (EMAIL_KEY, )

    def __init__(self, number, name, label, slots=None, image=None):
        pslots = {PHONE_KEY: number, NAME_KEY: name, LABEL_KEY: label}
        if slots:
            pslots.update(slots)
        ContactLeaf.__init__(self, pslots, name, image)

    def repr_key(self):
        return self.object[PHONE_KEY]

    def get_description(self):
        return '%s: %s' % (self.object[LABEL_KEY], self.object[PHONE_KEY])


class AddressContact(ContactLeaf):
    grouping_slots = ContactLeaf.grouping_slots + (EMAIL_KEY, )

    def __init__(self, address, name, label, slots=None, image=None):
        aslots = {ADDRESS_KEY: address, NAME_KEY: name, LABEL_KEY: label}
        if slots:
            aslots.update(slots)
        ContactLeaf.__init__(self, aslots, name, image)

    def repr_key(self):
        return self.object[ADDRESS_KEY]



import gettext
import locale
import sys

_debug = False

def setup_locale_and_gettext():
    """Set up localization with gettext"""
    package_name = "kupfer"
    localedir = "./locale"
    try:
        from kupfer import version_subst
    except ImportError:
        pass
    else:
        package_name = version_subst.PACKAGE_NAME
        localedir = version_subst.LOCALEDIR
    # Install _() builtin for gettext; always returning unicode objects
    # also install ngettext()
    gettext.install(package_name, localedir=localedir, #unicode=True,
            names=("ngettext",))
    # For Gtk.Builder, we need to call the C library gettext functions
    # As well as set the codeset to avoid locale-dependent translation
    # of the message catalog
    locale.bindtextdomain(package_name, localedir)
    locale.bind_textdomain_codeset(package_name, "UTF-8")
    # to load in current locale properly for sorting etc
    try:
        locale.setlocale(locale.LC_ALL, "")
    except locale.Error:
        pass

setup_locale_and_gettext()

def prt(*args):
    enc = locale.getpreferredencoding(do_setlocale=False)
    sys.stdout.buffer.write(" ".join(args).encode(enc, "replace"))
    sys.stdout.buffer.write(b"\n")
    #print(((" ".join(args)).encode(enc, "replace")))

def get_options():
    """Return a list of other application flags with --* prefix included."""

    program_options = [
        ("no-splash", _("do not present main interface on launch")),
        ("list-plugins", _("list available plugins")),
        ("debug", _("enable debug info")),
        # TRANS: --exec-helper=HELPER is an internal command
        # TRANS: that executes a helper program that is part of kupfer
        ("exec-helper=", _("run plugin helper")),
    ]
    misc_options = [
        ("help", _("show usage help")),
        ("version", _("show version information")),
    ]

    import getopt

    def make_help_text():
        usage_string = _("Usage: kupfer [ OPTIONS | FILE ... ]")
        def format_options(opts):
            return "\n".join("  --%-15s  %s" % (o,h) for o,h in opts)

        options_string = "%s\n\n%s\n\n%s\n" % (usage_string,
                format_options(program_options), format_options(misc_options))

        return options_string

    def make_plugin_list():
        from kupfer.core import plugins
        plugin_header = _("Available plugins:")
        plugin_list = plugins.get_plugin_desc()
        return "\n".join((plugin_header, plugin_list))

    # Fix sys.argv that can be None in exceptional cases
    if sys.argv[0] is None:
        sys.argv[0] = "kupfer"

    try:
        opts, args = getopt.getopt(sys.argv[1:], "",
                [o for o,h in program_options] +
                [o for o,h in misc_options])
    except getopt.GetoptError as exc:
        prt(str(exc))
        prt(make_help_text())
        raise SystemExit(1)

    for k, v in opts:
        if k == "--list-plugins":
            prt(gtkmain(make_plugin_list))
            raise SystemExit
        if k == "--help":
            prt(make_help_text())
            raise SystemExit
        if k == "--version":
            print_version()
            raise SystemExit
        if k == "--debug":
            global _debug
            _debug = True
        if k == "--relay":
            prt("WARNING: --relay is deprecated!")
            exec_helper('kupfer.keyrelay')
            raise SystemExit
        if k == "--exec-helper":
            exec_helper(v)
            raise SystemExit(1)

    # return list first of tuple pair
    return [tupl[0] for tupl in opts]

def print_version():
    from kupfer import version
    prt(version.PACKAGE_NAME, version.VERSION)

def print_banner():
    from kupfer import version

    if not sys.stdout or not sys.stdout.isatty():
        return

    banner = _(
        "%(PROGRAM_NAME)s: %(SHORT_DESCRIPTION)s\n"
        "   %(COPYRIGHT)s\n"
        "   %(WEBSITE)s\n") % vars(version)
    prt(banner)

def _set_process_title():
    try:
        import setproctitle
    except ImportError:
        pass
    else:
        setproctitle.setproctitle("kupfer")

def exec_helper(helpername):
    import runpy
    runpy.run_module(helpername, run_name='__main__', alter_sys=True)
    raise SystemExit

def gtkmain(run_function, *args, **kwargs):
    import gi
    gi.require_version("Gtk", "3.0")
    gi.require_version("Gdk", "3.0")

    return run_function(*args, **kwargs)

def browser_start(quiet):
    from gi.repository import Gdk
    if not Gdk.Screen.get_default():
        print("No Screen Found, Exiting...", file=sys.stderr)
        sys.exit(1)

    from kupfer.ui import browser
    w = browser.WindowController()
    w.main(quiet=quiet)

def main():
    # parse commandline before importing UI
    cli_opts = get_options()
    print_banner()

    from kupfer import pretty, version

    if _debug:
        pretty.debug = _debug
        pretty.print_debug(__name__, "Version:", version.PACKAGE_NAME, version.VERSION)
        try:
            import debug
            debug.install()
        except ImportError:
            pass
    sys.excepthook = sys.__excepthook__
    _set_process_title()

    quiet = ("--no-splash" in cli_opts)
    gtkmain(browser_start, quiet)




'''
Common objects for session_* plugins.
'''

from kupfer.objects import Source, RunnableLeaf
from kupfer import utils, pretty

__version__ = "2009-12-05"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"


def launch_argv_with_fallbacks(commands, print_error=True):
    """Try the sequence of @commands with utils.spawn_async,
    and return with the first successful command.
    return False if no command is successful and log an error
    """
    for argv in commands:
        ret = utils.spawn_async(argv)
        if ret: return ret
    pretty.print_error(__name__, "Unable to run command(s)", commands)
    return False

class CommandLeaf (RunnableLeaf):
    """The represented object of the CommandLeaf is a list of commandlines"""
    def run(self):
        launch_argv_with_fallbacks(self.object)

class Logout (CommandLeaf):
    """Log out from desktop"""
    def __init__(self, commands, name=None):
        if not name: name = _("Log Out...")
        CommandLeaf.__init__(self, commands, name)
    def get_description(self):
        return _("Log out or change user")
    def get_icon_name(self):
        return "system-log-out"

class Shutdown (CommandLeaf):
    """Shutdown computer or reboot"""
    def __init__(self, commands, name=None):
        if not name: name = _("Shut Down...")
        CommandLeaf.__init__(self, commands, name)
    def get_description(self):
        return _("Shut down, restart or suspend computer")
    def get_icon_name(self):
        return "system-shutdown"

class LockScreen (CommandLeaf):
    """Lock screen"""
    def __init__(self, commands, name=None):
        if not name: name = _("Lock Screen")
        CommandLeaf.__init__(self, commands, name)
    def get_description(self):
        return _("Enable screensaver and lock")
    def get_icon_name(self):
        return "system-lock-screen"

class CommonSource (Source):
    def __init__(self, name):
        super(CommonSource, self).__init__(name)
    def is_dynamic(self):
        return True
    def get_icon_name(self):
        return "system-shutdown"
    def provides(self):
        yield RunnableLeaf



# -*- coding: UTF-8 -*-

__kupfer_name__ = _("VirtualBox")
__kupfer_sources__ = ("VBoxMachinesSource", )
__description__ = _("Control VirtualBox Virtual Machines. "
                    "Supports both Sun VirtualBox and Open Source Edition.")
__version__ = "0.4"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

from kupfer.objects import Leaf, Action, Source
from kupfer import pretty
from kupfer import plugin_support
from kupfer.obj.apps import ApplicationSource

from kupfer.plugin.virtualbox import ose_support
from kupfer.plugin.virtualbox import constants as vbox_const


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "force_cli",
        "label": _("Force use CLI interface"),
        "type": bool,
        "value": False,
    },
)


def _get_vbox():
    if __kupfer_settings__['force_cli']:
        pretty.print_info(__name__, 'Using cli...')
        return ose_support
    try:
        from kupfer.plugin.virtualbox import vboxapi4_support
        pretty.print_info(__name__, 'Using vboxapi4...')
        return vboxapi4_support
    except ImportError:
        pass
    try:
        from kupfer.plugin.virtualbox import vboxapi_support
        pretty.print_info(__name__, 'Using vboxapi...')
        return vboxapi_support
    except ImportError:
        pass
    pretty.print_info(__name__, 'Using cli...')
    return ose_support


class _VBoxSupportProxy:
    VBOX = None

    def __getattr__(self, attr):
        if not self.VBOX:
            self.reload_settings()
        return getattr(self.VBOX, attr)

    def reload_settings(self):
        pretty.print_debug(__name__, '_VBoxSupportProxy.reloading...')
        self.unload_module()
        self.VBOX = _get_vbox()

    def unload_module(self):
        if not self.VBOX:
            return
        self.VBOX.unload()
        self.VBOX = None


vbox_support = _VBoxSupportProxy()


class VirtualMachine(Leaf):
    def __init__(self, obj, name, description):
        Leaf.__init__(self, obj, name)
        self.description = description

    def get_description(self):
        return self.description

    def get_icon_name(self):
        return vbox_support.ICON

    def get_actions(self):
        state = vbox_support.get_machine_state(self.object)
        if state == vbox_const.VM_STATE_POWEROFF:
            yield VMAction(_('Power On'), 'system-run',
                    vbox_const.VM_START_NORMAL)
            yield VMAction(_('Power On Headless'), 'system-run',
                    vbox_const.VM_START_HEADLESS, -5)
        elif state == vbox_const.VM_STATE_POWERON:
            yield VMAction(_('Send Power Off Signal'), 'system-shutdown',
                    vbox_const.VM_ACPI_POWEROFF, -5)
            yield VMAction(_('Pause'), 'pause', vbox_const.VM_PAUSE)
            yield VMAction(_('Reboot'), 'system-reboot',
                    vbox_const.VM_REBOOT, -10)
        elif state == vbox_const.VM_STATE_SAVED:
            yield VMAction(_('Power On'), 'system-run',
                    vbox_const.VM_START_NORMAL)
            yield VMAction(_('Power On Headless'), 'system-run',
                    vbox_const.VM_START_HEADLESS, -5)
        else:  # VM_STATE_PAUSED
            yield VMAction(_('Resume'), 'resume', vbox_const.VM_RESUME)

        if state in (vbox_const.VM_STATE_POWERON, vbox_const.VM_STATE_PAUSED):
            yield VMAction(_('Save State'), 'system-supsend',
                    vbox_const.VM_SAVE)
            yield VMAction(_('Power Off'), 'system-shutdown',
                    vbox_const.VM_POWEROFF, -10)


class VMAction(Action):
    def __init__(self, name, icon, command, rank_adjust=0):
        Action.__init__(self, name)
        self._icon = icon
        self.rank_adjust = rank_adjust
        self.command = command

    def get_icon_name(self):
        return self._icon

    def item_types(self):
        yield VirtualMachine

    def activate(self, leaf):
        vbox_support.vm_action(self.command, leaf.object)


class VBoxMachinesSource(ApplicationSource):
    appleaf_content_id = ("virtualbox-ose", "virtualbox")

    def __init__(self, name=_("VirtualBox Machines")):
        Source.__init__(self, name)

    def initialize(self):
        if vbox_support.MONITORED_DIRS:
            self.monitor_token = self.monitor_directories(
                    *vbox_support.MONITORED_DIRS)
        __kupfer_settings__.connect("plugin-setting-changed", self._setting_changed)

    def finalize(self):
        if vbox_support:
            vbox_support.unload_module()

    def is_dynamic(self):
        return vbox_support.IS_DYNAMIC

    def get_items(self):
        for machine_id, machine_name, machine_desc in vbox_support.get_machines():
            yield VirtualMachine(machine_id, machine_name, machine_desc)

    def get_description(self):
        return None

    def get_icon_name(self):
        return vbox_support.ICON

    def provides(self):
        yield VirtualMachine

    def _setting_changed(self, _setting, _key, _value):
        if vbox_support:
            vbox_support.reload_settings()



import fnmatch
import io
import pickle
import sys

class universalset (object):
    def __contains__(self, item):
        return True

class ConservativeUnpickler (pickle.Unpickler):
    """An Unpickler that refuses to import new modules

    >>> import pickle

    >>> import kupfer.objects
    >>> ConservativeUnpickler.loads(pickle.dumps(kupfer.objects.FileLeaf("A")))
    <builtin.FileLeaf A>

    >>> ConservativeUnpickler.loads(pickle.dumps(eval))
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing unsafe __builtin__.eval

    >>> import sys
    >>> import kupfer.obj.base
    >>> pdata = pickle.dumps(kupfer.obj.base.Leaf(1, "A"))
    >>> del sys.modules["kupfer.obj.base"]
    >>> ConservativeUnpickler.loads(pdata)
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing to load module kupfer.obj.base
    """
    safe_modules = {
        "builtins" : set(["set", "sum", "object"]),
        "copy_reg" : set(["_reconstructor"]),
        "kupfer.*" : universalset(),
    }
    @classmethod
    def is_safe_symbol(cls, module, name):
        for pattern in cls.safe_modules:
            if fnmatch.fnmatchcase(module, pattern):
                return name in cls.safe_modules[pattern]
        return False

    def find_class(self, module, name):
        if module not in sys.modules:
            raise pickle.UnpicklingError("Refusing to load module %s" % module)
        if not self.is_safe_symbol(module, name):
            raise pickle.UnpicklingError("Refusing unsafe %s.%s" % (module, name))
        return pickle.Unpickler.find_class(self, module, name)

    @classmethod
    def loads(cls, pickledata):
        unpickler = cls(io.BytesIO(pickledata))
        return unpickler.load()

class BasicUnpickler (ConservativeUnpickler):
    """An Unpickler that can only unpickle persistend ids and select builtins

    >>> import pickle

    >>> import kupfer.objects
    >>> BasicUnpickler.loads(pickle.dumps(kupfer.objects.FileLeaf("A")))
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing unsafe kupfer.obj.objects.FileLeaf
    """

    safe_modules = {
        "__builtin__" : set(["object"]),
        "copy_reg" : set(["_reconstructor"]),
        "kupfer.puid" : set(["SerializedObject"]),
    }

if __name__ == '__main__':
    import doctest
    doctest.testmod()



__kupfer_name__ = _("Clipboards")
__kupfer_sources__ = ("ClipboardSource", )
__kupfer_actions__ = ("ClearClipboards", )
__description__ = _("Recent clipboards and clipboard proxy objects")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

from collections import deque

from gi.repository import Gdk
from gi.repository import Gio
from gi.repository import Gtk

from kupfer.objects import Source, TextLeaf, Action, SourceLeaf
from kupfer.objects import FileLeaf
from kupfer.obj.compose import MultipleLeaf
from kupfer import plugin_support
from kupfer import kupferstring, pretty

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "max",
        "label": _("Number of recent clipboards to remember"),
        "type": int,
        "value": 10,
    },
    {
        "key" : "use_selection",
        "label": _("Include selected text in clipboard history"),
        "type": bool,
        "value": False,
    },
    {
        "key" : "sync_selection",
        "label": _("Copy selected text to primary clipboard"),
        "type": bool,
        "value": False,
    },
)

class SelectedText (TextLeaf):
    qf_id = "selectedtext"
    def __init__(self, text):
        TextLeaf.__init__(self, text, _('Selected Text'))

    def __repr__(self):
        return "<%s %s>" % (__name__, self.qf_id)

class ClipboardText (TextLeaf):
    def get_description(self):
        numlines = self.object.count("\n") + 1
        desc = self.get_first_text_line(self.object)

        return ngettext('Clipboard "%(desc)s"',
            'Clipboard with %(num)d lines "%(desc)s"',
            numlines) % {"num": numlines, "desc": desc }

class CurrentClipboardText (ClipboardText):
    qf_id = "clipboardtext"
    def __init__(self, text):
        ClipboardText.__init__(self, text, _('Clipboard Text'))

    def __repr__(self):
        return "<%s %s>" % (__name__, self.qf_id)

class CurrentClipboardFile (FileLeaf):
    "represents the *unique* current clipboard file"
    qf_id = "clipboardfile"
    def __init__(self, filepath):
        """@filepath is a filesystem byte string `str`"""
        FileLeaf.__init__(self, filepath, _('Clipboard File'))

    def __repr__(self):
        return "<%s %s>" % (__name__, self.qf_id)

class CurrentClipboardFiles (MultipleLeaf):
    "represents the *unique* current clipboard if there are many files"
    qf_id = "clipboardfile"
    def __init__(self, paths):
        files = [FileLeaf(path) for path in paths]
        MultipleLeaf.__init__(self, files, _("Clipboard Files"))

    def __repr__(self):
        return "<%s %s>" % (__name__, self.qf_id)


class ClearClipboards(Action):
    def __init__(self):
        Action.__init__(self, _("Clear"))

    def activate(self, leaf):
        leaf.object.clear()

    def item_types(self):
        yield SourceLeaf

    def valid_for_item(self, leaf):
        return isinstance(leaf.object, ClipboardSource)

    def get_description(self):
        return _("Remove all recent clipboards")

    def get_icon_name(self):
        return "edit-clear"


class ClipboardSource (Source):
    def __init__(self):
        Source.__init__(self, _("Clipboards"))
        self.clipboards = deque()

    def initialize(self):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        self._sig_id1 = clip.connect("owner-change", self._clipboard_changed)
        clip = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        self._sig_id2 = clip.connect("owner-change", self._clipboard_changed)
        self.clipboard_uris = []
        self.clipboard_text = None
        self.selected_text = None

    def finalize(self):
        clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clip.disconnect(self._sig_id1)
        clip = Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)
        clip.disconnect(self._sig_id2)
        self._sig_id1 = None
        self._sig_id2 = None
        self.clipboard_uris = []
        self.clipboard_text = None
        self.selected_text = None
        self.mark_for_update()

    def _clipboard_changed(self, clip, event, *args):
        is_selection = (event.selection == Gdk.SELECTION_PRIMARY)
        clip.request_text(self._on_text_for_change, is_selection)
        if not is_selection:
            clip.request_uris(self._on_uris_for_change)

    def _on_text_for_change(self, clip, text, is_selection):
        if text is None:
            return

        max_len = __kupfer_settings__["max"]
        is_valid = bool(text and text.strip())
        is_sync_selection = (is_selection and
                             __kupfer_settings__["sync_selection"])

        if not is_selection or __kupfer_settings__["use_selection"]:
            if is_valid:
                self._add_to_history(text, is_selection)

        if is_sync_selection and is_valid:
            Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD).set_text(text, -1)

        if is_selection:
            self.selected_text = text
        if not is_selection or is_sync_selection:
            self.clipboard_text = text
        self._prune_to_length(max_len)
        self.mark_for_update()

    def _on_uris_for_change(self, clip, uris):
        self.clipboard_uris = uris if uris is not None else []
        self.mark_for_update()

    def _add_to_history(self, cliptext, is_selection):
        if cliptext in self.clipboards:
            self.clipboards.remove(cliptext)
        # if the previous text is a prefix of the new selection, supercede it
        if (is_selection and self.clipboards
                and (cliptext.startswith(self.clipboards[-1])
                or cliptext.endswith(self.clipboards[-1]))):
            self.clipboards.pop()
        self.clipboards.append(cliptext)

    def _prune_to_length(self, max_len):
        while len(self.clipboards) > max_len:
            self.clipboards.popleft()

    def get_items(self):
        # selected text
        if self.selected_text:
            yield SelectedText(self.selected_text)

        # produce the current clipboard files if any
        paths = [_f for _f in [Gio.File.new_for_uri(uri).get_path() for uri in self.clipboard_uris] if _f]
        if len(paths) == 1:
            yield CurrentClipboardFile(paths[0])
        if len(paths) > 1:
            yield CurrentClipboardFiles(paths)

        # put out the current clipboard text
        if self.clipboard_text:
            yield CurrentClipboardText(self.clipboard_text)
        # put out the clipboard history
        for t in reversed(self.clipboards):
            yield ClipboardText(t)

    def get_description(self):
        return __description__

    def get_icon_name(self):
        return "edit-paste"

    def provides(self):
        yield TextLeaf
        yield FileLeaf
        yield MultipleLeaf

    def clear(self):
        self.clipboards.clear()
        self.mark_for_update()



__kupfer_name__ = _("GNU Screen")
__kupfer_sources__ = ("ScreenSessionsSource", )
__description__ = _("Active GNU Screen sessions")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os

from kupfer.objects import Leaf, Action, Source
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer import utils



def screen_sessions_infos():
    """
    Yield tuples of pid, name, time, status
    for running screen sessions
    """
    pipe = os.popen("screen -list")
    output = pipe.read()
    for line in output.splitlines():
        fields = line.split("\t")
        if len(fields) == 4:
            empty, pidname, time, status = fields
            pid, name = pidname.split(".", 1)
            time = time.strip("()")
            status = status.strip("()")
            yield (pid, name, time, status)

def get_username():
    """Return username for current user"""
    import pwd
    info = pwd.getpwuid(os.geteuid())
    return info[0]

class ScreenSession (Leaf):
    """Represented object is the session pid as string"""
    def get_actions(self):
        return (AttachScreen(),)

    def is_valid(self):
        for pid, name, time, status in screen_sessions_infos():
            if self.object == pid:
                return True
        else:
            return False

    def get_description(self):
        for pid, name, time, status in screen_sessions_infos():
            if self.object == pid:
                break
        else:
            return "%s (%s)" % (self.name, self.object)
        # Handle localization of status
        status_dict = {
            "Attached": _("Attached"),
            "Detached": _("Detached"),
        }
        status = status_dict.get(status, status)
        return (_("%(status)s session (%(pid)s) created %(time)s") %
                {"status": status, "pid": pid, "time": time})

    def get_icon_name(self):
        return "gnome-window-manager"

class ScreenSessionsSource (Source, FilesystemWatchMixin):
    """Source for GNU Screen sessions"""
    def __init__(self):
        super(ScreenSessionsSource, self).__init__(_("Screen Sessions"))

    def initialize(self):
        ## the screen dir might not exist when we start
        ## luckily, gio can monitor directories before they exist
        self.screen_dir = (os.getenv("SCREENDIR") or
                "/var/run/screen/S-%s" % get_username())
        if not os.path.exists(self.screen_dir):
            self.output_debug("Screen socket dir or SCREENDIR not found")
        self.monitor_token = self.monitor_directories(self.screen_dir,
                                                      force=True)

    def get_items(self):
        if not os.path.exists(self.screen_dir):
            return
        for pid, name, time, status in screen_sessions_infos():
            yield ScreenSession(pid, name)

    def get_description(self):
        return _("Active GNU Screen sessions")
    def get_icon_name(self):
        return "terminal"
    def provides(self):
        yield ScreenSession

class AttachScreen (Action):
    """
    """
    def __init__(self):
        name = _("Attach")
        super(AttachScreen, self).__init__(name)
    def activate(self, leaf):
        pid = leaf.object
        action_argv = ['screen', '-x', '-R', ('%s' % pid)]
        utils.spawn_in_terminal(action_argv)




# -*- coding: UTF-8 -*-
__kupfer_name__ = _("Claws Mail")
__kupfer_sources__ = ("ClawsContactsSource", )
__kupfer_actions__ = ("NewMailAction", "SendFileByMail")
__description__ = _("Claws Mail Contacts and Actions")
__version__ = "2018-10-07"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os
from xml.dom import minidom
import xml

from kupfer.objects import Action
from kupfer.objects import TextLeaf, UrlLeaf, RunnableLeaf, FileLeaf
from kupfer import utils
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.contacts import ContactLeaf, EmailContact, email_from_leaf


class ComposeMail(RunnableLeaf):
    ''' Create new mail without recipient '''
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Compose New Email"))

    def run(self):
        utils.spawn_async(['claws-mail', '--compose'])

    def get_description(self):
        return _("Compose a new message in Claws Mail")

    def get_icon_name(self):
        return "mail-message-new"


class ReceiveMail(RunnableLeaf):
    ''' Receive all new mail from all accounts '''
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Receive All Email"))

    def run(self):
        utils.spawn_async(['claws-mail', '--receive-all'])

    def get_description(self):
        return _("Receive new messages from all accounts in ClawsMail")

    def get_icon_name(self):
        return "mail-send-receive"


class NewMailAction(Action):
    ''' Create new mail to selected leaf'''
    def __init__(self):
        Action.__init__(self, _('Compose Email'))

    def activate(self, leaf):
        self.activate_multiple((leaf, ))

    def activate_multiple(self, objects):
        recipients = ",".join(email_from_leaf(L) for L in objects)
        utils.spawn_async(["claws-mail", "--compose", recipients])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_for_item(self, item):
        return bool(email_from_leaf(item))


class SendFileByMail (Action):
    '''Create new e-mail and attach selected file'''
    def __init__(self):
        Action.__init__(self, _('Send in Email To...'))

    def activate(self, obj, iobj):
        self.activate_multiple((obj, ), (iobj, ))

    def activate_multiple(self, objects, iobjects):
        recipients = ",".join(email_from_leaf(I) for I in iobjects)
        attachlist = ["--attach"] + [L.object for L in objects]
        utils.spawn_async(["claws-mail", "--compose", recipients] + attachlist)

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, item):
        return not item.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_object(self, iobj, for_item=None):
        return bool(email_from_leaf(iobj))

    def get_description(self):
        return _("Compose new message in Claws Mail and attach file")

    def get_icon_name(self):
        return "document-send"


class ClawsContactsSource(AppLeafContentMixin, ToplevelGroupingSource,
                          FilesystemWatchMixin):
    appleaf_content_id = 'claws-mail'

    def __init__(self, name=_("Claws Mail Address Book")):
        super(ClawsContactsSource, self).__init__(name, "Contacts")
        self._claws_addrbook_dir = os.path.expanduser('~/.claws-mail/addrbook')
        self._claws_addrbook_index = os.path.join(self._claws_addrbook_dir,
                                                  "addrbook--index.xml")
        self._version = 4

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        if not os.path.isdir(self._claws_addrbook_dir):
            return

        self.monitor_token = self.monitor_directories(self._claws_addrbook_dir)

    def monitor_include_file(self, gfile):
        # monitor only addrbook-*.xml files
        return gfile and gfile.get_basename().endswith('.xml') and \
                gfile.get_basename().startswith("addrbook-")

    def get_items(self):
        if os.path.isfile(self._claws_addrbook_index):
            for addrbook_file in self._load_address_books():
                addrbook_filepath = os.path.join(self._claws_addrbook_dir,
                                                 addrbook_file)
                if not os.path.exists(addrbook_filepath):
                    continue

                try:
                    dtree = minidom.parse(addrbook_filepath)
                    persons = dtree.getElementsByTagName('person')
                    for person in persons:
                        cn = person.getAttribute('cn')
                        addresses = person.getElementsByTagName('address')
                        for address in addresses:
                            email = address.getAttribute('email')
                            yield EmailContact(email, cn)
                except (Exception, xml.parsers.expat.ExpatError) as err:
                    self.output_error(err)

        yield ComposeMail()
        yield ReceiveMail()

    def should_sort_lexically(self):
        # since it is a grouping source, grouping and non-grouping will be
        # separate and only grouping leaves will be sorted
        return True

    def get_description(self):
        return _("Contacts from Claws Mail Address Book")

    def get_icon_name(self):
        return "claws-mail"

    def provides(self):
        yield RunnableLeaf
        yield ContactLeaf

    def _load_address_books(self):
        ''' load list of address-book files '''
        try:
            dtree = minidom.parse(self._claws_addrbook_index)
            for book in dtree.getElementsByTagName('book'):
                yield book.getAttribute('file')
        except (Exception, xml.parsers.expat.ExpatError) as err:
            self.output_error(err)



__kupfer_name__ = _("Locate Files")
__kupfer_actions__ = (
        "Locate",
    )
__description__ = _("Search filesystem using locate")
__version__ = ""
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import subprocess

from kupfer.objects import Action, Source
from kupfer.objects import TextLeaf
from kupfer import icons, plugin_support
from kupfer import kupferstring
from kupfer.obj.objects import ConstructFileLeaf


__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "ignore_case",
        "label": _("Ignore case distinctions when searching files"),
        "type": bool,
        "value": True,
    },
)

class Locate (Action):
    def __init__(self):
        Action.__init__(self, _("Locate Files"))

    def is_factory(self):
        return True
    def activate(self, leaf):
        return LocateQuerySource(leaf.object)
    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search filesystem using locate")
    def get_gicon(self):
        return icons.ComposedIcon("gnome-terminal", self.get_icon_name())
    def get_icon_name(self):
        return "edit-find"

class LocateQuerySource (Source):
    def __init__(self, query):
        Source.__init__(self, name=_('Results for "%s"') % query)
        self.query = query
        self.max_items = 500

    def repr_key(self):
        return self.query

    def get_items(self):
        ignore_case = '--ignore-case' if __kupfer_settings__["ignore_case"] else ''
        # Start two processes, one to take the first hits, one
        # to take the remaining up to maximum. We start both at the same time
        # (regrettably, locate wont output streamingly to stdout)
        # but we ask the second for results only after iterating the first few
        first_num = 12
        first_command = ("locate --null --limit %d %s '%s'" %
                (first_num, ignore_case, self.query))
        full_command = ("locate --null --limit %d %s '%s'" %
                (self.max_items, ignore_case, self.query))
        p1 = subprocess.Popen(first_command, shell=True, stdout=subprocess.PIPE)
        p2 = subprocess.Popen(full_command, shell=True, stdout=subprocess.PIPE)

        def get_locate_output(proc, offset=0):
            out, ignored_err = proc.communicate()
            return (ConstructFileLeaf(kupferstring.fromlocale(f))
                    for f in out.split(b'\x00')[offset:-1])

        for F in get_locate_output(p1, 0):
            yield F
        for F in get_locate_output(p2, first_num):
            yield F

    def get_gicon(self):
        return icons.ComposedIcon("gnome-terminal", self.get_icon_name())
    def get_icon_name(self):
        return "edit-find"



"""
Access functions of Kupfer's Interface
"""
from gi.repository import Gtk

from kupfer import utils, version

def _get_time(ctxenv):
    return ctxenv.get_timestamp() if ctxenv else \
            Gtk.get_current_event_time()

def show_help(ctxenv=None):
    """
    Show Kupfer help pages, if possible
    """
    if not utils.show_help_url("help:%s" % version.PACKAGE_NAME):
        utils.show_url(version.HELP_WEBSITE)

_about_dialog = None

def show_about_dialog(ctxenv=None):
    """
    create an about dialog and show it
    """
    # Use only one instance, stored in _about_dialog
    global _about_dialog
    if _about_dialog:
        ab = _about_dialog
    else:
        ab = Gtk.AboutDialog()
        ab.set_program_name(version.PROGRAM_NAME)
        ab.set_icon_name(version.ICON_NAME)
        ab.set_logo_icon_name(version.ICON_NAME)
        ab.set_version(version.VERSION)
        ab.set_comments(version.SHORT_DESCRIPTION)
        ab.set_copyright(version.COPYRIGHT)
        ab.set_website(version.WEBSITE)
        ab.set_license(version.LICENSE)
        ab.set_authors(version.AUTHORS)
        if version.DOCUMENTERS:
            ab.set_documenters(version.DOCUMENTERS)
        if version.TRANSLATOR_CREDITS:
            ab.set_translator_credits(version.TRANSLATOR_CREDITS)
        if version.ARTISTS:
            ab.set_artists(version.ARTISTS)

        ab.connect("response", _response_callback)
        # do not delete window on close
        ab.connect("delete-event", lambda *ign: True)
        _about_dialog = ab
    if ctxenv:
        ctxenv.present_window(ab)
    else:
        ab.present()

def _response_callback(dialog, response_id):
    dialog.hide()


def show_preferences(ctxenv):
    from kupfer.ui import preferences
    win = preferences.GetPreferencesWindowController()
    if ctxenv:
        win.show_on_screen(ctxenv.get_timestamp(),
                           ctxenv.get_screen())
    else:
        win.show(_get_time(ctxenv))

def show_plugin_info(plugin_id, ctxenv=None):
    from kupfer.ui import preferences
    prefs = preferences.GetPreferencesWindowController()
    prefs.show_focus_plugin(plugin_id, _get_time(ctxenv))




import json

# Action Accelerator configuration
from kupfer import config
from kupfer import pretty

_repr_key = repr

class AccelConfig(pretty.OutputMixin):
    def __init__(self):
        self.accels = {}
        self.loaded = False
        self.changed = False

    def _filename(self):
        ret = config.save_config_file("action_accels.json")
        if ret is None:
            self.output_error("Can't find XDG_CONFIG_HOME")
        return ret

    def load(self, validate_func):
        if self.loaded:
            return True
        self.loaded = True
        data_file = self._filename()
        if data_file is None:
            return False

        try:
            with open(data_file, "r") as fp:
                self.accels = json.load(fp)
            self.output_debug("Read", data_file)
        except FileNotFoundError:
            return
        except Exception as exc:
            self.output_error("Failed to read:", data_file)
            self.output_exc()
            return

        try:
            self._valid_accel(validate_func)
        except:
            self.output_exc()
            self.accels = {}
        self.output_debug("Loaded", self.accels)

    def _valid_accel(self, validate_func):
        if not isinstance(self.accels, dict):
            raise TypeError("Accelerators must be a dictionary")
        self.accels = {str(k): str(v) for k, v in self.accels.items()}
        delete = set()
        for obj, k in self.accels.items():
            if not validate_func(k):
                delete.add(obj)
                self.output_error("Ignoring invalid accel", k, "for", obj)
        for obj in delete:
            self.accels.pop(obj, None)

    def get(self, obj):
        """
        Return accel key for @obj or None
        """
        return self.accels.get(_repr_key(obj))

    def set(self, obj, key):
        self.output_debug("Set", key, "for", _repr_key(obj))
        assert hasattr(obj, "activate")
        assert isinstance(key, str)
        self.accels[_repr_key(obj)] = key
        self.changed = True

    def store(self):
        if not self.changed:
            return
        data_file = self._filename()
        if data_file is None:
            return False
        self.output_debug("Writing to", data_file)
        try:
            with open(data_file, "w") as fp:
                json.dump(self.accels, fp, indent=4, sort_keys=True)
        except Exception as exc:
            self.output_error("Failed to write:", data_file)
            self.output_exc()
        self.changed = False



__kupfer_name__ = _("Dictionary")
__kupfer_actions__ = ("LookUp", )
__description__ = _("Look up word in dictionary")
__version__ = ""
__author__ = "Ulrik"

from kupfer.objects import Action, TextLeaf, OperationError
from kupfer import utils
from kupfer import plugin_support

dictionaries = {
    'gnome-dictionary': ['gnome-dictionary', '--look-up='],
    'mate-dictionary': ['mate-dictionary', '--look-up='],
    'purple': ['purple', '--define='],
    'xfce4-dict': ['xfce4-dict', '--dict', ''],
}

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "dictionary",
        "label": _("Dictionary"),
        "type": str,
        "alternatives": list(dictionaries.keys()),
        "value": 'gnome-dictionary',
    }
)

class LookUp (Action):
    def __init__(self):
        Action.__init__(self, _("Look Up"))
    def activate(self, leaf):
        text = leaf.object
        dict_id = __kupfer_settings__["dictionary"]
        dict_argv = list(dictionaries[dict_id])
        dict_argv[-1] = dict_argv[-1] + text
        try:
            utils.spawn_async_notify_as(dict_id + ".desktop", dict_argv)
        except utils.SpawnError as exc:
            raise OperationError(exc)

    def item_types(self):
        yield TextLeaf
    def valid_for_item(self, leaf):
        text = leaf.object
        return len(text.split("\n", 1)) <= 1
    def get_description(self):
        return _("Look up word in dictionary")
    def get_icon_name(self):
        return "accessories-dictionary"




import functools
import os
from kupfer import pretty, config

@functools.lru_cache(maxsize=1)
def is_kwin():
    """ Try to figure out if KWin is the current window manager.  """
    global Wnck

    try:
        from gi.repository import Wnck
    except ImportError:
        pass
    else:
        screen = Wnck.Screen.get_default()
        if screen is not None:
            wm = screen.get_window_manager_name()
            pretty.print_debug(__name__, "window manager is", wm)
            if wm:
                return wm.lower() == "kwin"
    return _desktop_environment_guess().lower() == "kde"

def _desktop_environment_guess():
    ret = os.getenv("XDG_CURRENT_DESKTOP") or ""
    pretty.print_debug(__name__, "desktop environment is", ret)
    return ret

def is_wayland():
    return bool(os.getenv("WAYLAND_DISPLAY"))

def allows_keybinder():
    return config.has_capability("KEYBINDER") and not is_wayland()



from gi.repository import Gio, GdkPixbuf, Gtk
from gi.repository.GLib import GError
from gi.repository.Gio import Icon, ThemedIcon, FileIcon, File
from gi.repository.Gio import FILE_ATTRIBUTE_STANDARD_ICON, FILE_ATTRIBUTE_THUMBNAIL_PATH

from kupfer import config
from kupfer import datatools
from kupfer import pretty
from kupfer import scheduler
from kupfer.kupferstring import tounicode

icon_cache = {}

# number of elements in icon lru cache (per icon size)
ICON_CACHE_SIZE = 15

LARGE_SZ = 128
SMALL_SZ = 24

## default fallbacks for our themable icons
kupfer_icon_fallbacks = {
    'kupfer-execute': 'system-run',
    'kupfer-object': 'text-x-generic',
    'kupfer-object-multiple': 'folder',
    'kupfer-catalog': 'system-search',
    'kupfer-search': 'system-search',
}

kupfer_locally_installed_names = set()

def _icon_theme_changed(theme):
    pretty.print_info(__name__, "Icon theme changed, clearing cache")
    global icon_cache
    icon_cache = {}

_default_theme = Gtk.IconTheme.get_default()
_default_theme.connect("changed", _icon_theme_changed)
_local_theme = Gtk.IconTheme.new()
_local_theme.set_search_path([])

def parse_load_icon_list(icon_list_data, get_data_func, plugin_name=None):
    """
    @icon_list_data: A bytestring whose lines identify icons
    @get_data_func: A function to return the data for a relative filename
    @plugin_name: plugin id, if applicable
    """
    try:
        icon_list_string = icon_list_data.decode("utf-8")
    except UnicodeDecodeError as exc:
        pretty.print_error(__name__, "Malformed icon-list in plugin %r: %r"
                          % (plugin_name, exc))
        raise
    for line in icon_list_string.splitlines():
        # ignore '#'-comments
        if line.startswith("#") or not line.strip():
            continue
        fields = list(map(str.strip, line.split('\t')))
        if len(fields) < 2:
            pretty.print_error(__name__, "Malformed icon-list line %r from %r" %
                               (line, plugin_name))
            continue
        icon_name, basename = fields[:2]
        override = ('!override' in fields)
        def wrap_get_data():
            return get_data_func(basename)
        load_icon_from_func(plugin_name, icon_name, wrap_get_data, override)

def load_icon_from_func(plugin_name, icon_name, get_data_func, override=False):
    """
    Load icon from @icon_data into the name @icon_name

    @get_data_func: function to retrieve the data if needed
    @override: override the icon theme
    """
    icon_name = tounicode(icon_name)
    if not override and icon_name in kupfer_locally_installed_names:
        pretty.print_debug(__name__, "Skipping existing", icon_name)
        return
    if not override and _default_theme.has_icon(icon_name):
        pretty.print_debug(__name__, "Skipping themed icon", icon_name)
        return
    try:
        icon_data = get_data_func()
    except:
        pretty.print_error(__name__, "Error loading icon %r for %r" %
                           (icon_name, plugin_name))
        pretty.print_exc(__name__)
        return
    for size in (SMALL_SZ, LARGE_SZ):
        pixbuf = get_pixbuf_from_data(icon_data, size, size)
        Gtk.IconTheme.add_builtin_icon(icon_name, size, pixbuf)
        pretty.print_debug(__name__, "Loading icon", icon_name, "at", size,
                "for", plugin_name)
    kupfer_locally_installed_names.add(icon_name)

def get_icon(key, icon_size):
    """
    try retrieve icon in cache
    is a generator so it can be concisely called with a for loop
    """
    try:
        rec = icon_cache[icon_size][key]
    except KeyError:
        return
    yield rec

def store_icon(key, icon_size, icon):
    """
    Store an icon in cache. It must not have been stored before
    """
    assert icon, "icon %s may not be %s" % (key, icon)
    icon_rec = icon
    if icon_size not in icon_cache:
        icon_cache[icon_size] = datatools.LruCache(ICON_CACHE_SIZE)
    icon_cache[icon_size][key] = icon_rec

def _get_icon_dwim(icon, icon_size):
    """Make an icon at @icon_size where
    @icon can be either an icon name, or a gicon
    """
    if isinstance(icon, Icon):
        return get_icon_for_gicon(icon, icon_size)
    elif icon:
        return get_icon_for_name(icon, icon_size)
    return None

class ComposedIcon (object):
    """
    A composed icon, which kupfer will render to pixbuf as
    background icon with the decorating icon as emblem
    """

    def __init__(self, baseicon, emblem, emblem_is_fallback=False):
        self.minimum_icon_size = 48
        self.baseicon = baseicon
        self.emblemicon = emblem

    @classmethod
    def new(cls, *args, **kwargs):
        """Contstuct a composed icon from @baseicon and @emblem,
        which may be GIcons or icon names (strings)
        """
        return cls(*args, **kwargs)


def ComposedIconSmall(baseicon, emblem, **kwargs):
    """Create composed icon for leaves with emblem visible on browser list"""
    ci = ComposedIcon(baseicon, emblem, **kwargs)
    ci.minimum_icon_size = SMALL_SZ
    return ci


def _render_composed_icon(composed_icon, icon_size):
    # If it's too small, render as fallback icon
    if icon_size < composed_icon.minimum_icon_size:
        return _get_icon_for_standard_gicon(composed_icon.baseicon, icon_size)
    emblemicon = composed_icon.emblemicon
    baseicon = composed_icon.baseicon
    toppbuf = _get_icon_dwim(emblemicon, icon_size)
    bottompbuf = _get_icon_dwim(baseicon, icon_size)
    if not toppbuf or not bottompbuf:
        return None

    dest = bottompbuf.copy()
    # @fr is the scale
    fr = 0.6
    dcoord = int((1-fr)*icon_size)
    dsize = int(fr*icon_size)
    # http://library.gnome.org/devel/gdk-pixbuf/unstable//gdk-pixbuf-scaling.html
    toppbuf.composite(dest, dcoord, dcoord, dsize, dsize,
            dcoord, dcoord, fr, fr, GdkPixbuf.InterpType.BILINEAR, 255)
    return dest

def get_thumbnail_for_gfile(gfile, width=-1, height=-1):
    """
    Return a Pixbuf thumbnail for the file at
    gfile: Gio.File
    size is @width x @height

    return None if not found
    """
    if not gfile.query_exists():
        return None
    finfo = gfile.query_info(FILE_ATTRIBUTE_THUMBNAIL_PATH, Gio.FileQueryInfoFlags.NONE, None)
    thumb_path = finfo.get_attribute_byte_string(FILE_ATTRIBUTE_THUMBNAIL_PATH)

    return get_pixbuf_from_file(thumb_path, width, height)

def get_pixbuf_from_file(thumb_path, width=-1, height=-1):
    """
    Return a Pixbuf thumbnail for the file at @thumb_path
    sized @width x @height
    For non-icon pixbufs:
    We might cache these, but on different terms than the icon cache
    if @thumb_path is None, return None
    """
    if not thumb_path:
        return None
    try:
        icon = GdkPixbuf.Pixbuf.new_from_file_at_size(thumb_path, width, height)
        return icon
    except GError as e:
        # this error is not important, the program continues on fine,
        # so we put it in debug output.
        pretty.print_debug(__name__, "get_pixbuf_from_file file:", thumb_path,
            "error:", e)

def get_gicon_for_file(uri):
    """
    Return a GIcon representing the file at
    the @uri, which can be *either* and uri or a path

    return None if not found
    """

    gfile = File.new_for_path(uri)
    if not gfile.query_exists():
        gfile = File.new_for_uri(uri)
        if not gfile.query_exists():
            return None

    finfo = gfile.query_info(FILE_ATTRIBUTE_STANDARD_ICON, Gio.FileQueryInfoFlags.NONE, None)
    gicon = finfo.get_attribute_object(FILE_ATTRIBUTE_STANDARD_ICON)
    return gicon

def get_icon_for_gicon(gicon, icon_size):
    """
    Return a pixbuf of @icon_size for the @gicon

    NOTE: Currently only the following can be rendered:
        Gio.ThemedIcon
        Gio.FileIcon
        kupfer.icons.ComposedIcon
    """
    # FIXME: We can't load any general GIcon
    if not gicon:
        return None
    if isinstance(gicon, ComposedIcon):
        return _render_composed_icon(gicon, icon_size)
    return _get_icon_for_standard_gicon(gicon, icon_size)

def _get_icon_for_standard_gicon(gicon, icon_size):
    """Render ThemedIcon and FileIcon"""
    if isinstance(gicon, FileIcon):
        ifile = gicon.get_file()
        return get_icon_from_file(ifile.get_path(), icon_size)
    if isinstance(gicon, ThemedIcon):
        names = gicon.get_names()
        return get_icon_for_name(names[0], icon_size, names)
    pretty.print_debug(__name__, "get_icon_for_gicon, could not load", gicon)
    return None


def _setup_icon_renderer(sched):
    from kupfer.core import settings
    setctl = settings.GetSettingsController()
    setctl.connect("alternatives-changed::icon_renderer", _icon_render_change)
    setctl.connect("value-changed::tools.icon_renderer", _icon_render_change)
    _icon_render_change(setctl)

def _icon_render_change(setctl, *arguments):
    global _IconRenderer
    renderer_dict = setctl.get_preferred_alternative('icon_renderer')
    renderer = renderer_dict.get("renderer")
    if not renderer or renderer is _IconRenderer:
        return
    pretty.print_debug(__name__, "Using", renderer)
    _icon_theme_changed(None)
    _IconRenderer = renderer

scheduler.GetScheduler().connect("loaded", _setup_icon_renderer)


class IconRenderer (object):
    """
    Default GTK+ implementation
    """
    @classmethod
    def pixbuf_for_name(cls, icon_name, icon_size):
        if icon_name in kupfer_locally_installed_names:
            try:
                return _local_theme.load_icon(icon_name, icon_size,
                                              Gtk.IconLookupFlags.USE_BUILTIN |
                                              Gtk.IconLookupFlags.FORCE_SIZE)
            except GError:
                pass
        try:
            return _default_theme.load_icon(icon_name, icon_size,
                                            Gtk.IconLookupFlags.USE_BUILTIN |
                                            Gtk.IconLookupFlags.FORCE_SIZE)
        except GError:
            pass

    @classmethod
    def pixbuf_for_file(cls, file_path, icon_size):
        try:
            icon = GdkPixbuf.Pixbuf.new_from_file_at_size(file_path, icon_size,
                                                        icon_size)
            return icon
        except GError:
            pretty.print_exc(__name__)

_IconRenderer = IconRenderer


def get_icon_for_name(icon_name, icon_size, icon_names=[]):
    for i in get_icon(icon_name, icon_size):
        return i
    if not icon_names: icon_names = (icon_name,)

    # Try the whole list of given names
    for load_name in icon_names:
        try:
            icon = _IconRenderer.pixbuf_for_name(load_name, icon_size)
            if icon:
                break
            elif icon_name in kupfer_icon_fallbacks:
                fallback_name = kupfer_icon_fallbacks[icon_name]
                icon = _IconRenderer.pixbuf_for_name(fallback_name, icon_size)
                if icon:
                    break
        except Exception:
            pretty.print_exc(__name__)
            icon = None
    else:
        # if we did not reach 'break' in the loop
        return None
    # We store the first icon in the list, even if the match
    # found was later in the chain
    store_icon(icon_name, icon_size, icon)
    return icon

def get_icon_from_file(icon_file, icon_size):
    # try to load from cache
    for icon in get_icon(icon_file, icon_size):
        return icon
    icon = _IconRenderer.pixbuf_for_file(icon_file, icon_size)
    if icon is not None:
        store_icon(icon_file, icon_size, icon)
        return icon

def is_good(gicon):
    """Return True if it is likely that @gicon will load a visible icon
    (icon name exists in theme, or icon references existing file)
    """
    if not gicon:
        return False
    if isinstance(gicon, ThemedIcon):
        return bool(get_good_name_for_icon_names(gicon.get_names()))
    if isinstance(gicon, FileIcon):
        ifile = gicon.get_file()
        return ifile.query_exists()
    # Since we can't load it otherwise (right now, see above)
    return False

def get_gicon_with_fallbacks(gicon, names):
    if not is_good(gicon):
        for name in names:
            gicon = ThemedIcon.new(name)
            if is_good(gicon):
                return gicon
        return ThemedIcon.new(name)
    return gicon

def get_good_name_for_icon_names(names):
    """Return first name in @names that exists
    in current icon theme, or None
    """
    for name in names:
        if _default_theme.has_icon(name):
            return name
    return None

def get_gicon_for_names(*names):
    return ThemedIcon.new_from_names(names)


def get_pixbuf_from_data(data, width=None, height=None):
    """Create pixbuf object from data with optional scaling

    @data: picture as raw data
    @width, @heigh: optional destination size
    """
    def set_size(img, img_width, img_height):
        scale = min(width/float(img_width), height/float(img_height))
        new_width, new_height = int(img_width*scale), int(img_height*scale)
        img.set_size(new_width, new_height)

    ploader = GdkPixbuf.PixbufLoader.new()
    if width and height:
        ploader.connect("size-prepared", set_size)
    ploader.write(data)
    ploader.close()
    return ploader.get_pixbuf()




from kupfer import datatools
from kupfer import icons
from kupfer import pretty
from kupfer.utils import locale_sort
from kupfer.kupferstring import tounicode, tofolded

__all__ = [
    "Error", "InvalidDataError", "OperationError", "InvalidLeafError",
    "KupferObject", "Leaf", "Action", "Source", "TextSource",
]

# If no gettext function is loaded at this point, we load a substitute,
# so that testing code can still work
import builtins
if not hasattr(builtins, "_"):
    def identity(x): return x
    builtins._ = identity

class Error (Exception):
    pass

class InvalidDataError (Error):
    "The data is wrong for the given Leaf"

class OperationError (Error):
    "Command execution experienced an error"

class InvalidLeafError (OperationError):
    "The Leaf passed to an Action is invalid"

_builtin_modules = frozenset([
    "kupfer.obj.objects",
    "kupfer.obj.base",
    "kupfer.obj.sources",
    "kupfer.obj.fileactions",
])

class _BuiltinObject (type):
    def __new__(mcls, name, bases, dict):
        dict["_is_builtin"] = dict["__module__"] in _builtin_modules
        return type.__new__(mcls, name, bases, dict)


class KupferObject (object, metaclass=_BuiltinObject):
    """
    Base class for kupfer data model

    This class provides a way to get at an object's:

    * icon with get_thumbnail, get_pixbuf and get_icon
    * name with unicode() or str()
    * description with get_description

    @rank_adjust should be used _very_ sparingly:
        Default actions should have +5 or +1
        Destructive (dangerous) actions should have -5 or -10

    @fallback_icon_name is a class attribute for the last fallback
    icon; it must always be accessible.
    """
    rank_adjust = 0
    fallback_icon_name = "kupfer-object"
    def __init__(self, name=None):
        """ Init kupfer object with, where
        @name *should* be a unicode object but *may* be
        a UTF-8 encoded `str`
        """
        if not name:
            name = self.__class__.__name__
        self.name = tounicode(name)
        folded_name = tofolded(self.name)
        self.kupfer_add_alias(folded_name)

    def kupfer_add_alias(self, alias):
        if alias != str(self):
            if not hasattr(self, "name_aliases"):
                self.name_aliases = set()
            self.name_aliases.add(alias)

    def __str__(self):
        return self.name

    def __repr__(self):
        key = self.repr_key()
        keys = " %s" % (key, ) if (key is not None and key != "") else ""
        if self._is_builtin:
            return "<builtin.%s%s>" % (self.__class__.__name__, keys)
        return "<%s.%s%s>" % (self.__module__, self.__class__.__name__, keys)

    def repr_key(self):
        """
        Return an object whose str() will be used in the __repr__,
        self is returned by default.
        This value is used to recognize objects, for example learning commonly
        used objects.
        """
        return self

    def get_description(self):
        """Return a description of the specific item
        which *should* be a unicode object
        """
        return None

    def get_thumbnail(self, width, height):
        """Return pixbuf of size @width x @height if available
        Most objects will not implement this
        """
        return None

    def get_pixbuf(self, icon_size):
        """
        Returns an icon in pixbuf format with dimension @icon_size

        Subclasses should implement: get_gicon and get_icon_name,
        if they make sense.
        The methods are tried in that order.
        """
        gicon = self.get_gicon()
        pbuf = gicon and icons.get_icon_for_gicon(gicon, icon_size)
        if pbuf:
            return pbuf
        icon_name = self.get_icon_name()
        icon = icon_name and icons.get_icon_for_name(icon_name, icon_size)
        if icon:
            return icon
        return icons.get_icon_for_name(self.fallback_icon_name, icon_size)

    def get_icon(self):
        """
        Returns an icon in GIcon format

        Subclasses should implement: get_gicon and get_icon_name,
        if they make sense.
        The methods are tried in that order.
        """
        gicon = self.get_gicon()
        if gicon and icons.is_good(gicon):
            return gicon
        icon_name = self.get_icon_name()
        if icon_name and icons.get_good_name_for_icon_names((icon_name, )):
            return icons.get_gicon_for_names(icon_name)
        return icons.get_gicon_for_names(self.fallback_icon_name)

    def get_gicon(self):
        """Return GIcon, if there is one"""
        return None

    def get_icon_name(self):
        """Return icon name. All items should have at least
        a generic icon name to return.
        """
        return self.fallback_icon_name

def aslist(seq):
    """Return a list out of @seq, or seq if it is a list"""
    if not isinstance(seq, type([])) and not isinstance(seq, type(())):
        seq = list(seq)
    return seq

class _NonpersistentToken (object):
    "Goes None when pickled"
    __slots__ = "object"
    def __init__(self, object_):
        self.object = object_
    def __bool__(self):
        return bool(self.object)
    def __reduce__(self):
        return (sum, ((), None))

class Leaf (KupferObject):
    """
    Base class for objects

    Leaf.object is the represented object (data)
    All Leaves should be hashable (__hash__ and __eq__)
    """
    def __init__(self, obj, name):
        """Represented object @obj and its @name"""
        super(Leaf, self).__init__(name)
        self.object = obj
        self._content_source = None

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return (type(self) == type(other) and self.object == other.object)

    def add_content(self, content):
        """Register content source @content with Leaf"""
        self._content_source = content and _NonpersistentToken(content)

    def has_content(self):
        return self._content_source

    def content_source(self, alternate=False):
        """Content of leaf. it MAY alter behavior with @alternate,
        as easter egg/extra mode"""
        return self._content_source and self._content_source.object

    def get_actions(self):
        """Default (builtin) actions for this Leaf"""
        return ()

class Action (KupferObject):
    '''
    Base class for all actions

    Implicit interface:
      valid_object will be called once for each (secondary) object
      to see if it applies. If it is not defined, all objects are
      assumed ok (within the other type/source constraints)

    def valid_object(self, obj, for_item):
        """Whether @obj is good for secondary obj,
        where @for_item is passed in as a hint for
        which it should be applied to
        """
        return True

    @action_accelerator: str or None
        Default single lowercase letter key to use for selecting the action
        quickly
    '''
    fallback_icon_name = "kupfer-execute"
    action_accelerator = None

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        return (type(self) == type(other) and repr(self) == repr(other) and
                str(self) == str(other))

    def repr_key(self):
        """by default, actions of one type are all the same"""
        return ""

    def activate(self, obj, iobj=None, ctx=None):
        """Use this action with @obj and @iobj

        @obj:  the direct object (Leaf)
        @iobj: the indirect object (Leaf), if ``self.requires_object``
               returns ``False``

        if ``self.wants_context`` returns ``True``, then the action
        also receives an execution context object as ``ctx``.

        Also, ``activate_multiple(self, objects, iobjects=None, ctx=None)``
        is called if it is defined and the action gets either
        multiple objects or iobjects.
        """
        pass

    def wants_context(self):
        """Return ``True`` if ``activate`` should receive the
        ActionExecutionContext as the keyword argument context

        Defaults to ``False`` in accordance with the old protocol
        """
        return False

    def is_factory(self):
        """Return whether action may return a result collection as a Source"""
        return False

    def has_result(self):
        """Return whether action may return a result item as a Leaf"""
        return False

    def is_async(self):
        """If this action runs asynchronously, return True.

        Then activate(..) must return an object from the kupfer.task module,
        which will be queued to run by Kupfer's task scheduler.
        """
        return False

    def item_types(self):
        """Yield types this action may apply to. This is used only
        when this action is specified in __kupfer_actions__ to "decorate"
        """
        return ()

    def valid_for_item(self, item):
        """Whether action can be used with exactly @item"""
        return True

    def requires_object(self):
        """If this action requires a secondary object
        to complete is action, return True
        """
        return False

    def object_source(self, for_item=None):
        """Source to use for object or None,
        to use the catalog (flat and filtered for @object_types)
        """
        return None

    def object_source_and_catalog(self, for_item):
        return False

    def object_types(self):
        """Yield types this action may use as indirect objects, if the action
        requrires it.
        """
        return ()

class Source (KupferObject, pretty.OutputMixin):
    """
    Source: Data provider for a kupfer browser

    All Sources should be hashable and treated as equal if
    their @repr are equal!

    @source_user_reloadable if True source get "Reload" action without
        debug mode.
    @source_prefer_sublevel if True, the source by default exports
        its contents in a subcatalog, not to the toplevel.
        NOTE: *Almost never* use this: let the user decide, default to toplevel.
    @source_use_cache if True, the source can be pickled to disk to save its
        cached items until the next time the launcher is started.
    """
    fallback_icon_name = "kupfer-object-multiple"
    source_user_reloadable = False
    source_prefer_sublevel = False
    source_use_cache = True

    def __init__(self, name):
        KupferObject.__init__(self, name)
        self.cached_items = None
        self._version = 1

    @property
    def version(self):
        """version is for pickling (save and restore from cache),
        subclasses should increase self._version when changing"""
        return self._version

    def __eq__(self, other):
        return (type(self) == type(other) and repr(self) == repr(other) and
                self.version == other.version)

    def __hash__(self ):
        return hash(repr(self))

    def toplevel_source(self):
        return self

    def initialize(self):
        """
        Called when a Source enters Kupfer's system for real

        This method is called at least once for any "real" Source. A Source
        must be able to return an icon name for get_icon_name as well as a
        description for get_description, even if this method was never called.
        """
        pass

    def finalize(self):
        """
        Called before a source is deactivated.
        """
        pass

    def repr_key(self):
        return ""

    def get_items(self):
        """
        Internal method to compute and return the needed items

        Subclasses should use this method to return a sequence or
        iterator to the leaves it contains
        """
        return []

    def get_items_forced(self):
        """
        Force compute and return items for source.
        Default - call get_items method.
        """
        return self.get_items()

    def is_dynamic(self):
        """
        Whether to recompute contents each time it is accessed
        """
        return False

    def mark_for_update(self):
        """
        Mark source as changed

        it should be reloaded on next used (if normally cached)
        """
        self.cached_items = None

    def should_sort_lexically(self):
        """
        Sources should return items by most relevant order (most
        relevant first). If this is True, Source will sort items
        from get_item() in locale lexical order
        """
        return False

    def get_leaves(self, force_update=False):
        """
        Return a list of all leaves.

        Subclasses should implement get_items, so that Source
        can handle sorting and caching.
        if @force_update, ignore cache, print number of items loaded
        """
        if self.should_sort_lexically():
            # sort in locale order
            sort_func = locale_sort
        else:
            sort_func = lambda x: x

        if self.is_dynamic():
            if force_update:
                return sort_func(self.get_items_forced())
            else:
                return sort_func(self.get_items())

        if self.cached_items is None or force_update:
            if force_update:
                self.cached_items = aslist(sort_func(self.get_items_forced()))
                self.output_debug("Loaded %d items" % len(self.cached_items))
            else:
                self.cached_items = \
                        datatools.SavedIterable(sort_func(self.get_items()))
                self.output_debug("Loaded items")
        return self.cached_items

    def has_parent(self):
        return False

    def get_parent(self):
        return None

    def get_leaf_repr(self):
        """Return, if appicable, another object
        to take the source's place as Leaf"""
        return None

    def provides(self):
        """A seq of the types of items it provides;
        empty is taken as anything -- however most sources
        should set this to exactly the type they yield
        """
        return ()

    def get_search_text(self):
        return _("Type to search")

    def get_empty_text(self):
        return _("%s is empty") % str(self)


class TextSource (KupferObject):
    """TextSource base class implementation,

    this is a psedo Source"""
    def __init__(self, name=None, placeholder=None):
        """
        name: Localized name
        placeholder: Localized placeholder when it has no input
        """
        if not name:
            name = _("Text")
        KupferObject.__init__(self, name)
        self.placeholder = placeholder

    def __eq__(self, other):
        return (type(self) == type(other) and repr(self).__eq__(repr(other)))

    def __hash__(self ):
        return hash(repr(self))

    def initialize(self):
        pass

    def repr_key(self):
        return ""

    def get_rank(self):
        """All items are given this rank"""
        return 20

    def get_items(self, text):
        return ()

    def get_text_items(self, text):
        """Get leaves for unicode string @text"""
        return self.get_items(text)

    def has_parent(self):
        return False

    def provides(self):
        """A seq of the types of items it provides"""
        yield Leaf

    def get_icon_name(self):
        return "edit-select-all"

    def get_search_text(self):
        return self.placeholder if self.placeholder else _("Text")

    def get_empty_text(self):
        return self.placeholder if self.placeholder else _("Text")


class ActionGenerator (object):
    """A "source" for actions

    NOTE: The ActionGenerator should not perform any expensive
    computation, and not access any slow media (files, network) when
    returning actions.  Such expensive checks must be performed in
    each Action's valid_for_item method.
    """

    def get_actions_for_leaf(self, leaf):
        '''Return actions appropriate for given leaf. '''
        return []



import sys

from gi.repository import GObject

keyring = None

from kupfer import pretty
from kupfer import config
from kupfer.core import settings
from kupfer.core import plugins
from kupfer import utils

__all__ = [
    "UserNamePassword",
    "PluginSettings",
    "check_dbus_connection",
    "check_keyring_support",
]

def _is_core_setting(key):
    return key.startswith("kupfer_")

class PluginSettings (GObject.GObject, pretty.OutputMixin):
    """Allows plugins to have preferences by assigning an instance
    of this class to the plugin's __kupfer_settings__ attribute.

    Setting values are accessed by the getitem operator [] with
    the setting's 'key' attribute

    Signals:

        plugin-setting-changed: key, value

    """
    __gtype_name__ = "PluginSettings"

    def __init__(self, *setdescs):
        """Create a settings collection by passing in dictionaries
        as arguments, where each dictionary must have the following keys:
            key
            type
            value (default value)
            label (localized label)

        the @key may be any string except strings starting with
        'kupfer_', which are reserved
        """
        GObject.GObject.__init__(self)
        self.setting_descriptions = {}
        self.setting_key_order = []
        self.signal_connection = -1
        req_keys = set(("key", "value", "type", "label"))
        for desc in setdescs:
            if not req_keys.issubset(list(desc.keys())):
                missing = req_keys.difference(list(desc.keys()))
                raise KeyError("Plugin setting missing keys: %s" % missing)
            if _is_core_setting(desc["key"]):
                raise KeyError("Reserved plugin setting key: %r" % (desc["key"], ))
            self.setting_descriptions[desc["key"]] = dict(desc)
            self.setting_key_order.append(desc["key"])

    def __iter__(self):
        return iter(self.setting_key_order)

    def initialize(self, plugin_name):
        """Init by reading from global settings and setting up callbacks"""
        setctl = settings.GetSettingsController()
        for key in self:
            value_type = self.setting_descriptions[key]["type"]
            value = setctl.get_plugin_config(plugin_name, key, value_type)
            if value is not None:
                self[key] = value
            elif _is_core_setting(key):
                default = self.setting_descriptions[key]["value"]
                setctl.set_plugin_config(plugin_name, key, default, value_type)
        setctl.connect("value-changed", self._value_changed, plugin_name)
        # register for unload notification
        if not plugin_name.startswith("core."):
            plugins.register_plugin_unimport_hook(plugin_name,
                    self._disconnect_all, plugin_name)

    def __getitem__(self, key):
        return self.setting_descriptions[key]["value"]
    def __setitem__(self, key, value):
        value_type = self.setting_descriptions[key]["type"]
        self.setting_descriptions[key]["value"] = value_type(value)
        if not _is_core_setting(key):
            self.emit("plugin-setting-changed::"+str(key), key, value)

    def _value_changed(self, setctl, section, key, value, plugin_name):
        """Preferences changed, update object"""
        if key in self and plugin_name in section:
            self[key] = value

    def get_value_type(self, key):
        """Return type of setting @key"""
        return self.setting_descriptions[key]["type"]
    def get_label(self, key):
        """Return label for setting @key"""
        return self.setting_descriptions[key]["label"]
    def get_alternatives(self, key):
        """Return alternatives for setting @key (if any)"""
        return self.setting_descriptions[key].get("alternatives")
    def get_tooltip(self, key):
        """Return tooltip string for setting @key (if any)"""
        return self.setting_descriptions[key].get("tooltip")

    def connect_settings_changed_cb(self, callback, *args):
        self.signal_connection = \
                self.connect("plugin-setting-changed", callback, *args)

    def _disconnect_all(self, plugin_name):
        if self.signal_connection != -1:
            self.disconnect(self.signal_connection)


# Arguments: Key, Value
# Detailed by the key
GObject.signal_new("plugin-setting-changed", PluginSettings,
        GObject.SignalFlags.RUN_LAST | GObject.SignalFlags.DETAILED,
        GObject.TYPE_BOOLEAN,
        (GObject.TYPE_STRING, GObject.TYPE_PYOBJECT))

# Plugin convenience functions for dependencies

_has_dbus_connection = None

def check_dbus_connection():
    """
    Check if a connection to the D-Bus daemon is available,
    else raise ImportError with an explanatory error message.

    For plugins that can not be used without contact with D-Bus;
    if this check is used, the plugin may use D-Bus and assume it
    is available in the Plugin's code.
    """
    global _has_dbus_connection
    if _has_dbus_connection is None:
        import dbus
        try:
            dbus.Bus()
            _has_dbus_connection = True
        except dbus.DBusException:
            _has_dbus_connection = False
    if not _has_dbus_connection:
        raise ImportError(_("No D-Bus connection to desktop session"))

class UserNamePassword (object):
    pass

def check_keyring_support():
    """
    raise ImportError with because it is not supported
    """
    raise ImportError("Keyring is not supported")

def check_keybinding_support():
    """
    Check if we can make global keybindings
    """
    from kupfer.ui import keybindings
    if not keybindings.is_available():
        raise ImportError(_("Dependency '%s' is not available") % "Keybinder")

def _plugin_configuration_error(plugin, err):
    pretty.print_error(__name__, err)


def _is_valid_terminal(term_dict):
    if len(term_dict["argv"]) < 1:
        return False
    exe = term_dict["argv"][0]
    return bool(utils.lookup_exec_path(exe))


_available_alternatives = {
    "terminal": {
        "filter": _is_valid_terminal,
        "required_keys": {
            'name': str,
            'argv': list,
            'exearg': str,
            'desktopid': str,
            'startup_notify': bool,
        },
    },
    "icon_renderer": {
        "filter": None,
        "required_keys": {
            'name': str,
            'renderer': object,
        },
    },
}

_alternatives = {
    "terminal": {},
    "icon_renderer": {},
}


def register_alternative(caller, category_key, id_, **kwargs):
    """
    Register a new alternative for the category @category_key

    @caller: Must be the caller's plugin id (Plugin __name__ variable)

    @id_ is a string identifier for the object to register
    @kwargs are the keyed arguments for the alternative constructor

    Returns True with success
    """
    caller = str(caller)
    category_key = str(category_key)
    id_ = str(id_)

    if category_key not in _available_alternatives:
        _plugin_configuration_error(caller,
                "Category '%s' does not exist" % category_key)
        return
    alt = _available_alternatives[category_key]
    id_ = caller + "." + id_
    kw_set = set(kwargs)
    req_set = set(alt["required_keys"])
    if not req_set.issubset(kw_set):
        _plugin_configuration_error(caller,
            "Configuration error for alternative '%s':" % category_key)
        _plugin_configuration_error(caller, "Missing keys: %s" %
                (req_set - kw_set))
        return
    _alternatives[category_key][id_] = kwargs
    pretty.print_debug(__name__,
        "Registered alternative %s: %s" % (category_key, id_))
    setctl = settings.GetSettingsController()
    setctl._update_alternatives(category_key, _alternatives[category_key],
                                alt["filter"])

    # register the alternative to be unloaded
    plugin_id = ".".join(caller.split(".")[2:])
    if plugin_id and not plugin_id.startswith("core."):
        plugins.register_plugin_unimport_hook(plugin_id,
                _unregister_alternative, caller, category_key, id_)
    return True

def _unregister_alternative(caller, category_key, full_id_):
    """
    Remove the alternative for category @category_key
    (this is done automatically at plugin unload)
    """
    if category_key not in _available_alternatives:
        _plugin_configuration_error(caller,
                "Category '%s' does not exist" % category_key)
        return
    alt = _available_alternatives[category_key]
    id_ = full_id_
    try:
        del _alternatives[category_key][id_]
    except KeyError:
        _plugin_configuration_error(caller,
                "Alternative '%s' does not exist" % (id_, ))
        return
    pretty.print_debug(__name__,
        "Unregistered alternative %s: %s" % (category_key, id_))
    setctl = settings.GetSettingsController()
    setctl._update_alternatives(category_key, _alternatives[category_key],
                                alt["filter"])
    return True


